"""
xminds.api.client
~~~~~~~~~~~~~~~~~

This module implements the requests for all API endpoints.
The client handles the logic to automatically get a new JWT token using the refresh token
"""

import base64
from functools import wraps
import logging
import sys
import time
from urllib.parse import quote
from binascii import Error as BinasciiError

import numpy

from ..compat import tqdm
from .apirequest import CrossingMindsApiJsonRequest, CrossingMindsApiPythonRequest
from .exceptions import DuplicatedError, JwtTokenExpired, ServerError


def require_login(method):
    @wraps(method)
    def wrapped(self, *args, **kwargs):
        try:
            return method(self, *args, **kwargs)
        except JwtTokenExpired:
            if not self._refresh_token or not self.auto_refresh_token:
                raise
            self.login_refresh_token()
            return method(self, *args, **kwargs)
    return wrapped


class CrossingMindsApiClient:

    def __init__(self, serializer='pkl', **api_kwargs):
        self._database = None
        self._refresh_token = None
        self.auto_refresh_token = True
        if serializer.lower() in ['pkl', 'pickle']:
            cls = CrossingMindsApiPythonRequest
            self.b64_encode_bytes = False
        elif serializer.lower() == 'json':
            cls = CrossingMindsApiJsonRequest
            self.b64_encode_bytes = True
        else:
            raise NotImplementedError(f'unknown serializer {serializer}')
        self.api = cls(**api_kwargs)

    # === Account ===

    @require_login
    def create_individual_account(self, first_name, last_name, email, password, role='backend'):
        """
        Create a new individual account

        :param str first_name:
        :param str last_name:
        :param str email:
        :param str password:
        :returns: {'id': str, 'warnings?': [str]}
        """
        path = f'accounts/individual/'
        data = {
            'first_name': first_name,
            'last_name': last_name,
            'email': email,
            'password': password,
            'role': role,
        }
        return self.api.post(path=path, data=data)

    @require_login
    def partial_update_individual_account(
            self, email, password=None, first_name=None, last_name=None):
        """
        Update the password or other editable property of an individual account,
        identified by an email.
        Editable properties not provided will not be modified.

        :param str email:
        :param str? password:
        :param str? first_name:
        :param str? last_name:
        :returns: {'warnings?': [str]}
        """
        path = f'accounts/individual/'
        data = {
            'email': email
        }
        if password:
            data['password'] = password
        if first_name:
            data['first_name'] = first_name,
        if last_name:
            data['last_name'] = last_name,
        return self.api.patch(path=path, data=data)

    @require_login
    def create_service_account(self, name, password, role='frontend'):
        """
        Create a new service account

        :param str name:
        :param str password:
        :param str? role:
        :returns: {'id': str, 'warnings?': [str]}
        """
        path = f'accounts/service/'
        data = {
            'name': name,
            'password': password,
            'role': role,
        }
        return self.api.post(path=path, data=data)

    @require_login
    def partial_update_service_account(self, name, password):
        """
        Update the password or other editable property of an existing service account,
        identified by a service name.
        Editable properties not provided will not be modified.

        :param str name:
        :param str password:
        :returns: {'warnings?': [str]}
        """
        path = f'accounts/service/'
        data = {
            'name': name,
            'password': password
        }
        return self.api.patch(path=path, data=data)

    def resend_verification_code(self, email):
        """
        Resend the verification code to the account email

        :param str email:
        """
        path = f'accounts/resend-verification-code/'
        return self.api.put(path=path, data={'email': email})

    def verify_account(self, code, email):
        """
        Verify the email of an account by entering the verification code

        :param str code:
        :param str email:
        :returns: {'verified': bool, 'email': str, 'warnings?': [str]}
        """
        path = f'accounts/verify/'
        data = {
            'code': code,
            'email': email,
        }
        return self.api.get(path=path, params=data)

    @require_login
    def list_accounts(self):
        """
        Get all accounts on the current organization

        :returns: {
            'individual_accounts': [
                {
                    'first_name': str,
                    'last_name': str,
                    'email': str,
                    'role': str,
                    'verified': bool,
                },
            ],
            'service_accounts': [
                {
                    'name': str,
                    'role': str,
                },
            ],
            'warnings?': [str]
        }
        """
        path = f'organizations/accounts/'
        return self.api.get(path=path)

    @require_login
    def delete_individual_account(self, email):
        """
        Delete an individual account

        :param str email:
        :raises: NotFoundError
        """
        path = 'accounts/individual/'
        data = {
            'email': email,
        }
        return self.api.delete(path=path, data=data)

    @require_login
    def delete_service_account(self, name):
        """
        Delete a service account

        :param str name: Service name
        :raises: NotFoundError
        """
        path = 'accounts/service/'
        data = {
            'name': name,
        }
        return self.api.delete(path=path, data=data)

    # === Login ===

    def login_individual(self, email, password, db_id, frontend_user_id=None,
                         frontend_session_id=None):
        """
        Login on a database with an account

        :param str email:
        :param str password:
        :param str db_id:
        :param ID? frontend_user_id: user ID
        :param ID? frontend_session_id: anonymous session ID
        :returns: {
            'token': str,
            'database': {
                'id': str,
                'name': str,
                'description': str,
                'item_id_type': str,
                'user_id_type': str,
                'session_id_type': str,
            },
            'warnings?': [str]
        }
        """
        path = f'login/individual/'
        data = {
            'email': email,
            'password': password,
            'db_id': db_id,
        }
        if frontend_user_id is not None:
            # cannot use `_userid2body` since we are not logged in yet
            if isinstance(frontend_user_id, bytes) and self.b64_encode_bytes:
                frontend_user_id = self._b64_encode(frontend_user_id)
            data['frontend_user_id'] = frontend_user_id
        if frontend_session_id is not None:
            # cannot use `_sessionid2body` since we are not logged in yet
            if isinstance(frontend_session_id, bytes) and self.b64_encode_bytes:
                frontend_session_id = self._b64_encode(frontend_session_id)
            data['frontend_session_id'] = frontend_session_id
        return self._login(path, data)

    def login_service(self, name, password, db_id, frontend_user_id=None,
                      frontend_session_id=None):
        """
        Login on a database with a service account

        :param str name:
        :param str password:
        :param str db_id:
        :param ID? frontend_user_id: user ID
        :param ID? frontend_session_id: anonymous session ID
        :returns: {
            'token': str,
            'database': {
                'id': str,
                'name': str,
                'description': str,
                'item_id_type': str,
                'user_id_type': str,
                'session_id_type': str,
            },
            'warnings?': [str]
        }
        """
        path = f'login/service/'
        data = {
            'name': name,
            'password': password,
            'db_id': db_id,
        }
        if frontend_user_id is not None:
            # cannot use `_userid2body` since we are not logged in yet
            if isinstance(frontend_user_id, bytes) and self.b64_encode_bytes:
                frontend_user_id = self._b64_encode(frontend_user_id)
            data['frontend_user_id'] = frontend_user_id
        if frontend_session_id is not None:
            # cannot use `_sessionid2body` since we are not logged in yet
            if isinstance(frontend_session_id, bytes) and self.b64_encode_bytes:
                frontend_session_id = self._b64_encode(frontend_session_id)
            data['frontend_session_id'] = frontend_session_id
        return self._login(path, data)

    def login_root(self, email, password):
        """
        Login with the root account without selecting a database

        :param str email:
        :param str password:
        :returns: {
            'token': str,
            'warnings?': [str]
        }
        """
        path = f'login/root/'
        data = {
            'email': email,
            'password': password
        }
        resp = self.api.post(path=path, data=data)
        jwt_token = resp['token']
        self.set_jwt_token(jwt_token)
        self._database = None
        self._refresh_token = None
        return resp

    def login_refresh_token(self, refresh_token=None):
        """
        Login again using a refresh token

        :param str? refresh_token: (default: self._refresh_token)
        :returns: {
            'token': str,
            'database': {
                'id': str,
                'name': str,
                'description': str,
                'item_id_type': str,
                'user_id_type': str,
                'session_id_type': str,
            },
            'warnings?': [str]
        }
        """
        refresh_token = refresh_token or self._refresh_token
        path = f'login/refresh-token/'
        data = {
            'refresh_token': refresh_token
        }
        return self._login(path, data)

    def _login(self, path, data):
        resp = self.api.post(path=path, data=data)
        jwt_token = resp['token']
        self.set_jwt_token(jwt_token)
        self._database = resp['database']
        self._refresh_token = resp['refresh_token']
        return resp

    # === Org metadata ===

    @require_login
    def get_organization(self):
        """
        Get organization meta-data
        :returns: {'metadata': dict, 'warnings?': [str]}
        """
        path = f'organizations/current/'
        return self.api.get(path=path)

    @require_login
    def create_or_partial_update_organization(self, metadata, preserve=None):
        """
        Create, or apply deep partial update of meta-data
        :param dict metadata: meta-data to store structured as unvalidated JSON-compatible dict
        :param bool? preserve: set to `True` to append values instead of replace as in RFC7396
        """
        path = f'organizations/current/'
        data = {'metadata': metadata}
        if preserve is not None:
            data['preserve'] = preserve
        return self.api.patch(path=path, data=data)

    @require_login
    def partial_update_organization_preview(self, metadata, preserve=None):
        """
        Preview deep partial update of extra data, without changing any state
        :param dict metadata: extra meta-data to store structured as unvalidated JSON-compatible dict
        :param bool? preserve: set to `True` to append values instead of replace as in RFC7396
        :returns: {
            'metadata_old': {
                'description': str,
                'extra': {**any-key: any-val},
            },
            'metadata_new': {
                'description': str,
                'extra': {**any-key: any-val},
            },
            'warnings?': [str]
        }
        """
        path = f'organizations/current/preview/'
        data = {'metadata': metadata}
        if preserve is not None:
            data['preserve'] = preserve
        return self.api.patch(path=path, data=data)

    # === Database ===

    @require_login
    def create_database(self, name, description, item_id_type, user_id_type, session_id_type=None,
                        item_filter_cache_expiration=None):
        """
        Create a new database

        :param str name: Database name, must be unique
        :param str description:
        :param str item_id_type: Item ID type
        :param str user_id_type: User ID type
        :param str? session_id_type: Anonymous Session ID type
        :param float? item_filter_cache_expiration: Refresh cache time of items filters (in seconds)
        :returns: {'id': str, 'warnings?': str}
        """
        path = f'databases/'
        data = {
            'name': name,
            'description': description,
            'item_id_type': item_id_type,
            'user_id_type': user_id_type,
        }
        if item_filter_cache_expiration:
            data['item_filter_cache_expiration'] = item_filter_cache_expiration
        if session_id_type:
            data['session_id_type'] = session_id_type
        return self.api.post(path=path, data=data)

    @require_login
    def get_all_databases(self, amt=None, page=None):
        """
        Get all databases on the current organization

        :param int? amt: amount of databases by page (default: API default)
        :param int? page: page number (default: 1)
        :returns: {
            'has_next': bool,
            'next_page': int,
            'databases': [
                {
                    'id': int,
                    'name': str,
                    'description': str,
                    'item_id_type': str,
                    'user_id_type': str,
                    'session_id_type': str,
                },
            ],
            'warnings?': [str]
        }
        """
        path = f'databases/'
        params = {}
        if amt:
            params['amt'] = amt
        if page:
            params['page'] = page
        return self.api.get(path=path, params=params)

    @require_login
    def get_database(self):
        """
        Get details on current database

        :returns: {
            'id': str,
            'name': str,
            'description': str,
            'item_id_type': str,
            'user_id_type': str,
            'session_id_type': str,
            'counters': {
                'rating': int,
                'user': int,
                'item': int,
            },
            'metadata': {**any-key: any-val},
            'warnings?': [str]
        }
        """
        path = f'databases/current/'
        return self.api.get(path=path)

    @require_login
    def partial_update_database(self, description=None, metadata=None, preserve=None):
        """
        update description, and apply deep partial update of extra meta-data
        :param str? description: description of DB
        :param dict? metadata: extra data to store structured as unvalidated JSON-compatible dict
        :param bool? preserve: set to `True` to append values instead of replace as in RFC7396
        """
        path = f'databases/current/'
        data = {}
        assert description is not None or metadata is not None
        if description is not None:
            data['description'] = description
        if metadata is not None:
            data['metadata'] = metadata
        if preserve is not None:
            data['preserve'] = preserve
        return self.api.patch(path=path, data=data)

    @require_login
    def partial_update_database_preview(self, description=None, metadata=None, preserve=None):
        """
        Preview deep partial update of extra data, without changing any state
        :param str? description: description of DB
        :param dict? metadata: extra data to store structured as unvalidated JSON-compatible dict
        :param bool? preserve: set to `True` to append values instead of replace as in RFC7396
        :returns: {
            'metadata_old': {
                'description': str,
                'metadata': {**any-key: any-val},
            },
            'metadata_new': {
                'description': str,
                'metadata': {**any-key: any-val},
            },
            'warnings?': [str]
        }
        """
        path = f'databases/current/preview/'
        data = {}
        assert description is not None or metadata is not None
        if description is not None:
            data['description'] = description
        if metadata is not None:
            data['metadata'] = metadata
        if preserve is not None:
            data['preserve'] = preserve
        return self.api.patch(path=path, data=data)

    @require_login
    def delete_database(self):
        """
        Delete current database.
        """
        path = f'databases/current/'
        return self.api.delete(path=path, timeout=29)

    @require_login
    def status(self):
        """
        Get readiness status of current database.
        :returns: {'status': str, 'warnings?': [str]}
        """
        path = f'databases/current/status/'
        return self.api.get(path=path)

    # === User Property ===

    @require_login
    def get_user_property(self, property_name):
        """
        Get one user-property.

        :param str property_name: property name
        :returns: {
            'property_name': str,
            'value_type': str,
            'repeated': bool,
            'warnings?': [str]
        }
        """
        path = f'users-properties/{self.escape_url(property_name)}/'
        return self.api.get(path=path)

    @require_login
    def list_user_properties(self):
        """
        Get all user-properties for the current database.

        :returns: {
            'properties': [{
                'property_name': str,
                'value_type': str,
                'repeated': bool,
            }],
            'warnings?': [str]
        }
        """
        path = f'users-properties/'
        return self.api.get(path=path)

    @require_login
    def create_user_property(self, property_name, value_type, repeated=False):
        """
        Create a new user-property.

        :param str property_name: property name
        :param str value_type: property type
        :param bool? repeated: whether the property has many values (default: False)
        """
        path = f'users-properties/'
        data = {
            'property_name': property_name,
            'value_type': value_type,
            'repeated': repeated,
        }
        return self.api.post(path=path, data=data)

    @require_login
    def delete_user_property(self, property_name):
        """
        Delete an user-property given by its name

        :param str property_name: property name
        """
        path = f'users-properties/{self.escape_url(property_name)}/'
        return self.api.delete(path=path)

    # === User ===

    @require_login
    def get_user(self, user_id):
        """
        Get one user given its ID.

        :param ID user_id: user ID
        :returns: {
            'item': {
                'id': ID,
                *<property_name: property_value>,
            },
            'warnings?': [str]
        }
        """
        user_id = self._userid2url(user_id)
        path = f'users/{user_id}/'
        resp = self.api.get(path=path)
        resp['user']['user_id'] = self._body2userid(resp['user']['user_id'])
        return resp

    @require_login
    def list_users(self, users_id):
        """
        Get multiple users given their IDs.
        The users in the response are not aligned with the input.
        In particular this endpoint does not raise NotFoundError if any user in missing.
        Instead, the missing users are simply not present in the response.

        :param ID-array users_id: users IDs
        :returns: {
            'users': array with fields ['id': ID, *<property_name: value_type>]
                contains only the non-repeated values,
            'users_m2m': dict of arrays for repeated values:
                {
                    *<repeated_property_name: {
                        'name': str,
                        'array': array with fields ['user_index': uint32, 'value_id': value_type],
                    }>
                },
            'warnings?': [str]
        }
        """
        users_id = self._userid2body(users_id)
        path = f'users-bulk/list/'
        data = {'users_id': users_id}
        resp = self.api.post(path=path, data=data)
        resp['users'] = self._body2itemid(resp['users'])
        return resp

    @require_login
    def list_users_paginated(self, amt=None, cursor=None):
        """
        Get multiple users by page.
        The response is paginated, you can control the response amount and offset
        using the query parameters ``amt`` and ``cursor``.

        :param int? amt: amount to return (default: use the API default)
        :param str? cursor: Pagination cursor
        :returns: {
            'users': array with fields ['id': ID, *<property_name: value_type>]
                contains only the non-repeated values,
            'users_m2m': dict of arrays for repeated values:
                {
                    *<repeated_property_name: {
                        'name': str,
                        'array': array with fields ['user_index': uint32, 'value_id': value_type],
                    }>
                },
            'has_next': bool,
            'next_cursor': str, pagination cursor to use in next request to get more users,
            'warnings?': [str],
        }
        """
        path = f'users-bulk/'
        params = {}
        if amt:
            params['amt'] = amt
        if cursor:
            params['cursor'] = cursor
        resp = self.api.get(path=path, params=params)
        resp['users'] = self._body2userid(resp['users'])
        return resp

    @require_login
    def create_or_update_user(self, user):
        """
        Create a new user, or update it if the ID already exists.

        :param dict user: user ID and properties {'user_id': ID, *<property_name: property_value>}
        """
        user = dict(user)
        user_id = self._userid2url(user.pop('user_id'))
        path = f'users/{user_id}/'
        data = {
            'user': user,
        }
        return self.api.put(path=path, data=data)

    @require_login
    def create_or_update_users_bulk(self, users, users_m2m=None, wait_for_completion=None,
                                    chunk_size=(1 << 10)):
        """
        Create many users in bulk, or update the ones for which the id already exist.

        :param array users: array with fields ['id': ID, *<property_name: value_type>]
            contains only the non-repeated values,
        :param dict? users_m2m: dict of arrays for repeated values:
            {
                *<repeated_property_name: {
                    'name': str,
                    'array': array with fields ['user_index': uint32, 'value_id': value_type],
                }>
            }
        :param int? chunk_size: split the requests in chunks of this size (default: 1K)
        :param bool? wait_for_completion: (default: True)
        """
        path = f'users-bulk/'
        for users_chunk, users_m2m_chunk in self._chunk_users(users, users_m2m, chunk_size):
            data = {
                'users': users_chunk,
                'users_m2m': users_m2m_chunk
            }
            if wait_for_completion is not None:
                data['wait_for_completion'] = wait_for_completion
            self.api.put(path=path, data=data, timeout=60)

    @require_login
    def partial_update_user(self, user, create_if_missing=None):
        """
        Partially update some properties of an user

        :param dict user: user ID and properties {'user_id': ID, *<property_name: property_value>}
        :param bool? create_if_missing: control whether an error should be returned or a new user
        should be created when the ``user_id`` does not already exist. (default: False)
        :returns: {
            'user_created': bool, True if the user did not exist and was created,
            'user_modified': bool True if the user already exists and was modified,
            'warnings?': [str],
        }
        """
        user = dict(user)
        user_id = self._userid2url(user.pop('user_id'))
        path = f'users/{user_id}/'
        data = {
            'user': user,
        }
        if create_if_missing is not None:
            data['create_if_missing'] = create_if_missing
        return self.api.patch(path=path, data=data)

    @require_login
    def partial_update_users_bulk(self, users, users_m2m=None, create_if_missing=None,
                                  chunk_size=(1 << 10)):
        """
        Partially update some properties of many users.

        :param array users: array with fields ['id': ID, *<property_name: value_type>]
            contains only the non-repeated values,
        :param dict? users_m2m: dict of arrays for repeated values:
            {
                *<repeated_property_name: {
                    'name': str,
                    'array': array with fields ['user_index': uint32, 'value_id': value_type],
                }>
            }
        :param bool? create_if_missing: to control whether an error should be returned or new users
            should be created when the ``user_id`` does not already exist. (default: False)
        :param int? chunk_size: split the requests in chunks of this size (default: 1K)
        """
        path = f'users-bulk/'
        data = {}
        if create_if_missing is not None:
            data['create_if_missing'] = create_if_missing
        for users_chunk, users_m2m_chunk in self._chunk_users(users, users_m2m, chunk_size):
            data['users'] = users_chunk
            data['users_m2m'] = users_m2m_chunk
            self.api.patch(path=path, data=data, timeout=60)

    @require_login
    def delete_user(self, user_id):
        """
        Delete a single user;  doesn't wait for task completion

        :param bytes user_id:
        """
        user_id_url = self._userid2url(user_id)
        self.api.delete(f'users/{user_id_url}')

    @require_login
    def delete_users(self, users_id):
        """
        Delete users; doesn't wait for task completion

        :param ID-array users_id: user IDs
        """
        data = {'users_id': self._userid2body(users_id)}
        self.api.delete(path='users-bulk/', data=data)

    def _chunk_users(self, users, users_m2m, chunk_size):
        users_m2m = users_m2m or []
        # cast dict to list of dict
        if isinstance(users_m2m, dict):
            users_m2m = [{'name': name, 'array': array}
                         for name, array in users_m2m.items()]
        n_chunks = int(numpy.ceil(len(users) / chunk_size))
        for i in tqdm(range(n_chunks), disable=(True if n_chunks < 4 else None)):
            start_idx = i * chunk_size
            end_idx = (i + 1) * chunk_size
            users_chunk = users[start_idx:end_idx]
            # split M2M array-optimized if any
            users_m2m_chunk = []
            for m2m in users_m2m:
                array = m2m['array']
                if isinstance(array, numpy.ndarray):
                    mask = (array['user_index'] >= start_idx) & (array['user_index'] < end_idx)
                    array_chunk = array[mask]  # does copy
                    array_chunk['user_index'] -= start_idx
                else:
                    logging.warning('array-optimized many-to-many format is not efficient '
                                    'with JSON. Use numpy arrays and pkl serializer instead')
                    array_chunk = [
                        {'user_index': row['user_index'] - start_idx, 'value_id': row['value_id']}
                        for row in array
                        if start_idx <= row['user_index'] < end_idx
                    ]
                users_m2m_chunk.append({'name': m2m['name'], 'array': array_chunk})
            yield self._userid2body(users_chunk), users_m2m_chunk

    # === Item Property ===

    @require_login
    def get_item_property(self, property_name):
        """
        Get one item-property.

        :param str property_name: property name
        :returns: {
            'property_name': str,
            'value_type': str,
            'repeated': bool,
            'warnings?': [str],
        }
        """
        path = f'items-properties/{self.escape_url(property_name)}/'
        return self.api.get(path=path)

    @require_login
    def list_item_properties(self):
        """
        Get all item-properties for the current database.

        :returns: {
            'properties': [{
                'property_name': str,
                'value_type': str,
                'repeated': bool,
            }],
            'warnings?': [str],
        }
        """
        path = f'items-properties/'
        return self.api.get(path=path)

    @require_login
    def create_item_property(self, property_name, value_type, repeated=False):
        """
        Create a new item-property.

        :param str property_name: property name
        :param str value_type: property type
        :param bool? repeated: whether the property has many values (default: False)
        """
        path = f'items-properties/'
        data = {
            'property_name': property_name,
            'value_type': value_type,
            'repeated': repeated,
        }
        return self.api.post(path=path, data=data)

    @require_login
    def delete_item_property(self, property_name):
        """
        Delete an item-property given by its name

        :param str property_name: property name
        """
        path = f'items-properties/{self.escape_url(property_name)}/'
        return self.api.delete(path=path)

    # === Item ===

    @require_login
    def get_item(self, item_id):
        """
        Get one item given its ID.

        :param ID item_id: item ID
        :returns: {
            'item': {
                'id': ID,
                *<property_name: property_value>,
            },
            'warnings?': [str],
        }
        """
        item_id = self._itemid2url(item_id)
        path = f'items/{item_id}/'
        resp = self.api.get(path=path)
        resp['item']['item_id'] = self._body2itemid(resp['item']['item_id'])
        return resp

    @require_login
    def list_items(self, items_id):
        """
        Get multiple items given their IDs.
        The items in the response are not aligned with the input.
        In particular this endpoint does not raise NotFoundError if any item in missing.
        Instead, the missing items are simply not present in the response.

        :param ID-array items_id: items IDs
        :returns: {
            'items': array with fields ['id': ID, *<property_name: value_type>]
                contains only the non-repeated values,
            'items_m2m': dict of arrays for repeated values:
                {
                    *<repeated_property_name: {
                        'name': str,
                        'array': array with fields ['item_index': uint32, 'value_id': value_type],
                    }>
                },
            'warnings?': [str],
        }
        """
        items_id = self._itemid2body(items_id)
        path = f'items-bulk/list/'
        data = {'items_id': items_id}
        resp = self.api.post(path=path, data=data)
        resp['items'] = self._body2itemid(resp['items'])
        return resp

    @require_login
    def list_items_paginated(self, amt=None, cursor=None):
        """
        Get multiple items by page.
        The response is paginated, you can control the response amount and offset
        using the query parameters ``amt`` and ``cursor``.

        :param int? amt: amount to return (default: use the API default)
        :param str? cursor: Pagination cursor
        :returns: {
            'items': array with fields ['id': ID, *<property_name: value_type>]
                contains only the non-repeated values,
            'items_m2m': dict of arrays for repeated values:
                {
                    *<repeated_property_name: {
                        'name': str,
                        'array': array with fields ['item_index': uint32, 'value_id': value_type],
                    }>
                },
            'has_next': bool,
            'next_cursor': str, pagination cursor to use in next request to get more items,
            'warnings?': [str],
        }
        """
        path = f'items-bulk/'
        params = {}
        if amt:
            params['amt'] = amt
        if cursor:
            params['cursor'] = cursor
        resp = self.api.get(path=path, params=params)
        resp['items'] = self._body2itemid(resp['items'])
        return resp

    @require_login
    def create_or_update_item(self, item, wait_for_completion=None):
        """
        Create a new item, or update it if the ID already exists.

        :param dict item: item ID and properties {'item_id': ID, *<property_name: property_value>}
        :param bool? wait_for_completion: (default: True)
        """
        item = dict(item)
        item_id = self._itemid2url(item.pop('item_id'))
        path = f'items/{item_id}/'
        data = {
            'item': item,
        }
        if wait_for_completion is not None:
            data['wait_for_completion'] = wait_for_completion
        return self.api.put(path=path, data=data)

    @require_login
    def create_or_update_items_bulk(self, items, items_m2m=None, wait_for_completion=None,
                                    chunk_size=(1 << 10)):
        """
        Create many items in bulk, or update the ones for which the id already exist.

        :param array items: array with fields ['id': ID, *<property_name: value_type>]
            contains only the non-repeated values,
        :param array? items_m2m: dict of arrays for repeated values:
            {
                *<repeated_property_name: {
                    'name': str,
                    'array': array with fields ['item_index': uint32, 'value_id': value_type],
                }>
            }
        :param bool? wait_for_completion: (default: True)
        :param int? chunk_size: split the requests in chunks of this size (default: 1000)
        """
        path = f'items-bulk/'
        for items_chunk, items_m2m_chunk in self._chunk_items(items, items_m2m, chunk_size):
            data = {
                'items': items_chunk,
                'items_m2m': items_m2m_chunk,
            }
            if wait_for_completion is not None:
                data['wait_for_completion'] = wait_for_completion
            self.api.put(path=path, data=data, timeout=60)

    @require_login
    def partial_update_item(self, item, create_if_missing=None):
        """
        Partially update some properties of an item.

        :param dict item: item ID and properties {'item_id': ID, *<property_name: property_value>}
        :param bool? create_if_missing: control whether an error should be returned or a new item
            should be created when the ``item_id`` does not already exist. (default: false)
        :returns: {
            'item_created': bool, True if the item did not exist and was created
            'item_modified': bool, True if the item already exists and was modified
            'warnings?': [str],
        }
        """
        item = dict(item)
        item_id = self._itemid2url(item.pop('item_id'))
        path = f'items/{item_id}/'
        data = {
            'item': item,
        }
        if create_if_missing is not None:
            data['create_if_missing'] = create_if_missing
        return self.api.patch(path=path, data=data)

    @require_login
    def partial_update_items_bulk(self, items, items_m2m=None, create_if_missing=None,
                                  chunk_size=(1 << 10), wait_for_completion=None):
        """
        Partially update some properties of many items.

        :param array items: array with fields ['id': ID, *<property_name: value_type>]
            contains only the non-repeated values,
        :param array? items_m2m: dict of arrays for repeated values:
            {
                *<repeated_property_name: {
                    'name': str,
                    'array': array with fields ['item_index': uint32, 'value_id': value_type],
                }>
            }
        :param bool? create_if_missing: control whether an error should be returned or a new item
        should be created when the ``item_id`` does not already exist. (default: False)
        :param int? chunk_size: split the requests in chunks of this size (default: 1000)
        :param bool? wait_for_completion: (default: True)
        """
        path = f'items-bulk/'
        data = {}
        if create_if_missing is not None:
            data['create_if_missing'] = create_if_missing
        if wait_for_completion is not None:
            data['wait_for_completion'] = wait_for_completion
        for items_chunk, items_m2m_chunk in self._chunk_items(items, items_m2m, chunk_size):
            data['items'] = items_chunk
            data['items_m2m'] = items_m2m_chunk
            self.api.patch(path=path, data=data, timeout=60)

    @require_login
    def delete_item(self, item_id):
        """
        Delete a single item;  doesn't wait for task completion

        :param bytes item_id:
        """
        item_id_url = self._itemid2url(item_id)
        self.api.delete(path=f'items/{item_id_url}/')

    @require_login
    def delete_items(self, items_id):
        """
        Delete items; doesn't wait for task completion

        :param ID-array items_id: items IDs
        """
        data = {'items_id': self._itemid2body(items_id)}
        self.api.delete(path='items-bulk/', data=data)

    def _chunk_items(self, items, items_m2m, chunk_size):
        items_m2m = items_m2m or []
        # cast dict to list of dict
        if isinstance(items_m2m, dict):
            items_m2m = [{'name': name, 'array': array}
                         for name, array in items_m2m.items()]
        n_chunks = int(numpy.ceil(len(items) / chunk_size))
        for i in tqdm(range(n_chunks), disable=(True if n_chunks < 4 else None)):
            start_idx = i * chunk_size
            end_idx = (i + 1) * chunk_size
            items_chunk = items[start_idx:end_idx]
            # split M2M array-optimized if any
            items_m2m_chunk = []
            for m2m in items_m2m:
                array = m2m['array']
                if isinstance(array, numpy.ndarray):
                    mask = (array['item_index'] >= start_idx) & (array['item_index'] < end_idx)
                    array_chunk = array[mask]  # does copy
                    array_chunk['item_index'] -= start_idx
                else:
                    logging.warning('array-optimized many-to-many format is not efficient '
                                    'with JSON. Use numpy arrays and pkl serializer instead')
                    array_chunk = [
                        {'item_index': row['item_index'] - start_idx, 'value_id': row['value_id']}
                        for row in array
                        if start_idx <= row['item_index'] < end_idx
                    ]
                items_m2m_chunk.append({'name': m2m['name'], 'array': array_chunk})
            yield self._itemid2body(items_chunk), items_m2m_chunk

    # === Reco: Item-to-item ===

    @require_login
    def get_reco_item_to_items(
            self, item_id, amt=None, cursor=None,
            scenario=None, filters=None, reranking=None,
            skip_default_scenario=None,
            user_id=None, session_id=None,
    ):
        """
        Get similar items.

        :param ID item_id: item ID
        :param int? amt: amount to return (default: use the API default)
        :param str? cursor: Pagination cursor
        :param str? scenario: name of scenario
        :param list-str? filters: Item-property filters. Filter format: ['<PROP_NAME>:<OPERATOR>:<OPTIONAL_VALUE>',...]
        :param list-str? reranking: Item-property reranking. Format: ['<PROP_NAME>:<OPERATOR>:<OPTIONAL_WEIGHT>:<OPTIONS>']
        :param bool? skip_default_scenario: True to skip default scenario if any
        :param ID? user_id: user ID. Only used in the context of an A/B test scenario to select the group A or B
            and keep track of the respective group in analytics, NOT used to personalize recommendations
        :param ID? session_id: session ID. Only used in the context of an A/B test scenario to select the group A or B
            and keep track of the respective group in analytics, NOT used to personalize recommendations
        :returns: {
            'items_id': array of items IDs,
            'next_cursor': str, pagination cursor to use in next request to get more items,
            'warnings?': [str],
            'evaluated_scenarios': {
                'runtime?': [{'scenario_type': str, 'to?': str, 'scenario_name': str}],
                'default?': [{'scenario_type': str, 'to?': str, 'scenario_name': str}]
            }
        }
        """
        item_id = self._itemid2url(item_id)
        path = f'recommendation/items/{item_id}/items/'
        params = {}
        if amt:
            params['amt'] = amt
        if cursor:
            params['cursor'] = cursor
        if filters:
            params['filters'] = self._clean_filters_to_urlsafe(filters)
        if reranking:
            params['reranking'] = self._clean_reranking_to_urlsafe(reranking)
        if scenario:
            params['scenario'] = scenario
        if skip_default_scenario is not None:
            params['skip_default_scenario'] = skip_default_scenario
        if user_id:
            params['user_id'] = user_id
        if session_id:
            params['session_id'] = session_id
        resp = self.api.get(path=path, params=params)
        resp['items_id'] = self._body2itemid(resp['items_id'])
        return resp

    @require_login
    def get_precomputed_reco_item_to_items(self, item_id, table=None, scenario=None,
                                           skip_default_scenario=None):
        """
        Get similar items, from a file of precomputed recommendations (stored as a ``table``).

        The table must be with ``LOCAL`` connection, have no compressions and have resource equal to
        ``ITEM_TO_ITEMS_RECO``.
        The table format must be indexed raw data file.

        :param ID item_id: item ID
        :param str? table: table name
        :param str? scenario: scenario name
        :param bool? skip_default_scenario: Specify whether default scenario should by applied or skipped
        :returns: {
            'items_id': array of items IDs,
            'warnings?': [str],
        }
        """
        item_id = self._itemid2url(item_id)
        path = f'recommendation/precomputed/items/{item_id}/items/'
        params = {}
        if table:
            params['table'] = table
        if scenario:
            params['scenario'] = scenario
        if skip_default_scenario is not None:
            params['skip_default_scenario'] = skip_default_scenario
        return self.api.get(path=path, params=params)

    # === Reco: Session-to-item ===

    @require_login
    def get_reco_session_to_items(self, ratings=None, interactions=None, user_properties=None,
                                  amt=None, cursor=None, scenario=None, filters=None,
                                  reranking=None, exclude_rated_items=None,
                                  skip_default_scenario=None, user_id=None, session_id=None):
        """
        Get items recommendations given the ratings or interactions of an anonymous session.
        Ratings and interactions are mutually exclusive.

        :param array? ratings: ratings array with fields ['item_id': ID, 'rating': float]
        :param array? interactions: interactions to calculate ratings from.
            Timestamp is optional and defaults to now.
            ['item_id': ID, 'interaction_type': 'O' 'timestamp?': float]
        :param dict? user_properties: user properties {**property_name: property_value(s)}
        :param int? amt: amount to return (default: use the API default)
        :param str? cursor: Pagination cursor
        :param str? scenario: scenario name
        :param list-str? filters: Item-property filters. Filter format: ['<PROP_NAME>:<OPERATOR>:<OPTIONAL_VALUE>',...]
        :param list-str? reranking: Item-property reranking. Format: ['<PROP_NAME>:<OPERATOR>:<OPTIONAL_WEIGHT>:<OPTIONS>']
        :param bool? exclude_rated_items: exclude rated items from response
        :param bool? skip_default_scenario: True to skip default scenario if any
        :param ID? user_id: user ID. Only used in the context of an A/B test scenario to select the group A or B
            and keep track of the respective group in analytics, NOT used to personalize recommendations
        :param ID? session_id: Anonymous Session ID.
        :returns: {
            'items_id': array of items IDs,
            'next_cursor': str, pagination cursor to use in next request to get more items,
            'warnings?': [str],
            'evaluated_scenarios': {
                'runtime?': [{'scenario_type': str, 'to?': str, 'scenario_name': str}],
                'default?': [{'scenario_type': str, 'to?': str, 'scenario_name': str}]
            }
        }
        """
        path = f'recommendation/sessions/items/'
        data = {}
        if ratings is not None:
            data['ratings'] = self._itemid2body(ratings)
        if user_properties:
            data['user_properties'] = user_properties
        if amt:
            data['amt'] = amt
        if cursor:
            data['cursor'] = cursor
        if filters:
            data['filters'] = filters
        if reranking:
            data['reranking'] = reranking
        if exclude_rated_items is not None:
            data['exclude_rated_items'] = exclude_rated_items
        if scenario:
            data['scenario'] = scenario
        if skip_default_scenario is not None:
            data['skip_default_scenario'] = skip_default_scenario
        if interactions is not None:
            data['interactions'] = interactions
        if user_id:
            data['user_id'] = user_id
        if session_id:
            data['session_id'] = session_id
        resp = self.api.post(path=path, data=data)
        resp['items_id'] = self._body2itemid(resp['items_id'])
        return resp

    # === Reco: User-to-item ===

    @require_login
    def get_reco_user_to_items(
            self, user_id, amt=None, cursor=None, scenario=None,
            filters=None, reranking=None,
            exclude_rated_items=None, skip_default_scenario=None,
    ):
        """
        Get items recommendations given a user ID.

        :param ID user_id: user ID
        :param int? amt: amount to return (default: use the API default)
        :param str? cursor: Pagination cursor
        :param str? scenario: scenario's name
        :param list-str? filters: Item-property filters. Filter format: ['<PROP_NAME>:<OPERATOR>:<OPTIONAL_VALUE>',...]
        :param list-str? reranking: Item-property reranking. Format: ['<PROP_NAME>:<OPERATOR>:<OPTIONAL_WEIGHT>:<OPTIONS>']
        :param bool? exclude_rated_items: exclude rated items from response
        :param bool? skip_default_scenario: True to skip default scenario if any
        :returns: {
            'items_id': array of items IDs,
            'next_cursor': str, pagination cursor to use in next request to get more items,
            'warnings?': [str],
            'evaluated_scenarios': {
                'runtime?': [{'scenario_type': str, 'to?': str, 'scenario_name': str}],
                'default?': [{'scenario_type': str, 'to?': str, 'scenario_name': str}]
            }
        }
        """
        user_id = self._userid2url(user_id)
        path = f'recommendation/users/{user_id}/items/'
        params = {}
        if amt:
            params['amt'] = amt
        if cursor:
            params['cursor'] = cursor
        if filters:
            params['filters'] = self._clean_filters_to_urlsafe(filters)
        if reranking:
            params['reranking'] = self._clean_reranking_to_urlsafe(reranking)
        if exclude_rated_items is not None:
            params['exclude_rated_items'] = exclude_rated_items
        if scenario:
            params['scenario'] = scenario
        if skip_default_scenario is not None:
            params['skip_default_scenario'] = skip_default_scenario
        resp = self.api.get(path=path, params=params)
        resp['items_id'] = self._body2itemid(resp['items_id'])
        return resp

    @require_login
    def get_precomputed_reco_user_to_items(self, user_id, table=None, scenario=None,
                                           skip_default_scenario=None):
        """
        Get item recommendations given a user profile, from a file of precomputed recommendations
        (stored as a ``table``).

        The table must be with ``LOCAL`` connection, have no compressions and have resource equal to
        ``USER_TO_ITEMS_RECO``.
        The table format must be indexed raw data file.

        :param ID user_id: user ID
        :param str? table: table name
        :param str? scenario: scenario name
        :param bool? skip_default_scenario: Specify whether default scenario should by applied or skipped
        :returns: {
            'items_id': array of items IDs,
            'warnings?': [str],
        }
        """
        user_id = self._userid2url(user_id)
        path = f'recommendation/precomputed/users/{user_id}/items/'
        params = {}
        if table:
            params['table'] = table
        if scenario:
            params['scenario'] = scenario
        if skip_default_scenario is not None:
            params['skip_default_scenario'] = skip_default_scenario
        return self.api.get(path=path, params=params)

    # === Reco: Group-users-to-item ===

    @require_login
    def get_reco_users_group_to_items(self, users, consensus_factor=None, amt=None, cursor=None,
                               scenario=None, filters=None, reranking=None,
                               skip_default_scenario=None):
        """
        Get items recommendations given a list of users IDs.

        :param array users: user array with fields ['user_id': ID]
        :param float? consensus_factor: consensus between the users' tastes
        :param int? amt: amount to return (default: use the API default)
        :param str? cursor: Pagination cursor
        :param str? scenario: scenario's name
        :param list-str? filters: Item-property filters. Filter format: ['<PROP_NAME>:<OPERATOR>:<OPTIONAL_VALUE>',...]
        :param list-str? reranking: Item-property reranking. Format: ['<PROP_NAME>:<OPERATOR>:<OPTIONAL_WEIGHT>:<OPTIONS>']
        :param bool? skip_default_scenario: True to skip default scenario if any
        :returns: {
            'items_id': array of items IDs,
            'next_cursor': str, pagination cursor to use in next request to get more items,
            'warnings?': [str],
        }
        """
        path = f'recommendation/users-groups/items/'
        data = {
            'users': users
        }
        if consensus_factor:
            data['consensus_factor'] = consensus_factor
        if amt:
            data['amt'] = amt
        if cursor:
            data['cursor'] = cursor
        if filters:
            data['filters'] = filters
        if reranking:
            data['reranking'] = reranking
        if scenario:
            data['scenario'] = scenario
        if skip_default_scenario is not None:
            data['skip_default_scenario'] = skip_default_scenario
        resp = self.api.post(path=path, data=data)
        resp['items_id'] = self._body2itemid(resp['items_id'])
        return resp

    # === Reco: User-to-item-property ===

    @require_login
    def get_reco_user_to_item_properties(
            self, user_id, item_property_name: str, amt=None,
            skip_default_scenario=None, filters=None, scenario=None):
        """
        Recommends item-property values given a user ID
        :param bytes user_id:
        :param str item_property_name:
        :param int? amt: (default 16)  maximal number of values to return from property `item_property_name`
        :param bool? skip_default_scenario: Specify whether default scenario should by applied or skipped
        :param list-str? filters: filters on intermediate items.
            Filter format: ['<PROP_NAME>:<OPERATOR>:<OPTIONAL_VALUE>',...]
        :param str? scenario:
        :raises: NotFoundError when data not found
        :raises: RequestError if property missing
        :return: {'properties': [n,] np.array, n<=amt, 'warnings?': [str]}
        """
        user_id = self._userid2url(user_id)
        path = f'recommendation/users/{user_id}/items-properties/{item_property_name}/'
        params = {}
        if amt:
            params['amt'] = amt
        if skip_default_scenario is not None:
            params['skip_default_scenario'] = skip_default_scenario
        if scenario:
            params['scenario'] = scenario
        if filters:
            params['filters'] = self._clean_filters_to_urlsafe(filters)
        resp = self.api.get(path=path, params=params)
        return resp

    # === Reco: Session-to-item-property ===

    @require_login
    def get_reco_session_to_item_properties(
            self, item_property_name: str,
            ratings=None,
            interactions=None,
            user_properties: dict = None,
            user_id=None, amt=None,
            skip_default_scenario=None, filters=None, scenario=None, session_id=None):
        """
        Recommends item-property values given a session
        :param str item_property_name:
        :param bytes? user_id: only used in scenario logic to select business rules
            in the context of an A/B test
        :param numpy.array[item_id,rating]? ratings:
        :param numpy.array[item_id,interaction_type,timestamp]? interactions:
        :param dict? user_properties:
        :param int? amt: (default 16) maximal number of values to return from property `item_property_name`
        :param bool? skip_default_scenario: Specify whether default scenario should by applied or skipped
        :param list-str? filters: filters on intermediate items.
            Filter format: ['<PROP_NAME>:<OPERATOR>:<OPTIONAL_VALUE>',...]
        :param str? scenario:
        :param bytes? session_id: Anonymous Session ID
        :raises: NotFoundError when data not found
        :raises: RequestError if property missing
        :return: {'properties': [n,] np.array, n<=amt, 'warnings?': [str]}
        """
        path = f'recommendation/sessions/items-properties/{item_property_name}/'
        data = {}
        if user_id:
            data['user_id'] = user_id
        if amt:
            data['amt'] = amt
        if skip_default_scenario is not None:
            data['skip_default_scenario'] = skip_default_scenario
        if scenario:
            data['scenario'] = scenario
        if filters:
            data['filters'] = self._clean_filters_to_urlsafe(filters)
        if ratings is not None:
            data['ratings'] = ratings
        if interactions is not None:
            data['interactions'] = interactions
        if user_properties:
            data['user_properties'] = user_properties
        if session_id:
            data['session_id'] = session_id
        resp = self.api.post(path=path, data=data)
        return resp

    # === Reco: Session-to-item with Context Items ===

    @require_login
    def get_reco_session_to_items_w_ctx_items(
            self, context_items, ratings=None, user_properties=None, amt=None, cursor=None,
            scenario=None, filters=None, reranking=None, exclude_rated_items=None,
            skip_default_scenario=None):
        """
        Get items recommendations given the ratings of an anonymous session and context items ID.

        :param array context_items: context items ID array with fields ['item_id': ID]
        :param array? ratings: ratings array with fields ['item_id': ID, 'rating': float]
        :param dict? user_properties: user properties {**property_name: property_value(s)}
        :param int? amt: amount to return (default: use the API default)
        :param str? cursor: Pagination cursor
        :param str? scenario: scenario name
        :param list-str? filters: Item-property filters. Filter format: ['<PROP_NAME>:<OPERATOR>:<OPTIONAL_VALUE>',...]
        :param list-str? reranking: Item-property reranking. Format: ['<PROP_NAME>:<OPERATOR>:<OPTIONAL_WEIGHT>:<OPTIONS>']
        :param bool? exclude_rated_items: exclude rated items from response
        :param bool? skip_default_scenario: True to skip default scenario if any
        :returns: {
            'items_id': array of items IDs,
            'next_cursor': str, pagination cursor to use in next request to get more items,
            'warnings?': [str],
        }
        """
        path = f'recommendation/context-items/sessions/items/'
        data = {
            'context_items': context_items,
        }
        if ratings is not None:
            data['ratings'] = self._itemid2body(ratings)
        if user_properties:
            data['user_properties'] = user_properties
        if amt:
            data['amt'] = amt
        if cursor:
            data['cursor'] = cursor
        if filters:
            data['filters'] = filters
        if reranking:
            data['reranking'] = reranking
        if exclude_rated_items is not None:
            data['exclude_rated_items'] = exclude_rated_items
        if scenario:
            data['scenario'] = scenario
        if skip_default_scenario is not None:
            data['skip_default_scenario'] = skip_default_scenario
        resp = self.api.post(path=path, data=data)
        resp['items_id'] = self._body2itemid(resp['items_id'])
        return resp

    # === Reco: User-to-item with Context Items ===

    @require_login
    def get_reco_user_to_items_w_ctx_items(
            self, context_items, user_id, amt=None, cursor=None, scenario=None, filters=None,
            reranking=None, exclude_rated_items=None, skip_default_scenario=None):
        """
        Get items recommendations given a user ID and context items ID.

        :param array context_items: context items ID array with fields ['item_id': ID]
        :param ID user_id: user ID
        :param int? amt: amount to return (default: use the API default)
        :param str? cursor: Pagination cursor
        :param str? scenario: scenario's name
        :param list-str? filters: Item-property filters. Filter format: ['<PROP_NAME>:<OPERATOR>:<OPTIONAL_VALUE>',...]
        :param list-str? reranking: Item-property reranking. Format: ['<PROP_NAME>:<OPERATOR>:<OPTIONAL_WEIGHT>:<OPTIONS>']
        :param bool? exclude_rated_items: exclude rated items from response
        :param bool? skip_default_scenario: True to skip default scenario if any
        :returns: {
            'items_id': array of items IDs,
            'next_cursor': str, pagination cursor to use in next request to get more items,
            'warnings?': [str],
        }
        """
        user_id = self._userid2url(user_id)
        path = f'recommendation/context-items/users/{user_id}/items/'
        data = {
            'context_items': context_items,
        }
        if amt:
            data['amt'] = amt
        if cursor:
            data['cursor'] = cursor
        if filters:
            data['filters'] = filters
        if reranking:
            data['reranking'] = reranking
        if exclude_rated_items is not None:
            data['exclude_rated_items'] = exclude_rated_items
        if scenario:
            data['scenario'] = scenario
        if skip_default_scenario is not None:
            data['skip_default_scenario'] = skip_default_scenario
        resp = self.api.post(path=path, data=data)
        resp['items_id'] = self._body2itemid(resp['items_id'])
        return resp

    # === User Ratings ===

    @require_login
    def create_or_update_rating(self, user_id, item_id, rating, timestamp=None):
        """
        Create or update a rating for a user and an item.
        If the rating exists for the tuple (user_id, item_id) then it is updated,
        otherwise it is created.

        :param ID user_id: user ID
        :param ID item_id: item ID
        :param float rating: rating value
        :param float? timestamp: rating timestamp (default: now)
        """
        user_id = self._userid2url(user_id)
        item_id = self._itemid2url(item_id)
        path = f'users/{user_id}/ratings/{item_id}/'
        data = {
            'rating': rating,
        }
        if timestamp is not None:
            data['timestamp'] = timestamp
        return self.api.put(path=path, data=data)

    @require_login
    def create_or_update_user_ratings_bulk(self, user_id, ratings):
        """
        Create or update bulks of ratings for a single user and many items.

        :param ID user_id: user ID
        :param array ratings: ratings array with fields:
            ['item_id': ID, 'rating': float, 'timestamp': float]
        """
        user_id = self._userid2url(user_id)
        path = f'users/{user_id}/ratings/'
        data = {
            'ratings': self._itemid2body(ratings),
        }
        return self.api.put(path=path, data=data, timeout=10)

    @require_login
    def create_or_update_ratings_bulk(self, ratings, chunk_size=(1 << 14)):
        """
        Create or update large bulks of ratings for many users and many items.

        :param array ratings: ratings array with fields:
            ['user_id': ID, 'item_id': ID, 'rating': float, 'timestamp': float]
        :param int? chunk_size: split the requests in chunks of this size (default: 16K)
        """
        path = f'ratings-bulk/'
        n_chunks = int(numpy.ceil(len(ratings) / chunk_size))
        sleep = chunk_size / 500
        for i in tqdm(range(n_chunks), disable=(True if n_chunks < 4 else None)):
            ratings_chunk = ratings[i*chunk_size:(i+1)*chunk_size]
            ratings_chunk = self._userid2body(self._itemid2body(ratings_chunk))
            data = {
                'ratings': ratings_chunk,
            }
            self.api.put(path=path, data=data, timeout=60)
            if n_chunks > 1:
                time.sleep(sleep)
        return

    @require_login
    def list_user_ratings(self, user_id, amt=None, page=None):
        """
        List the ratings of one user (paginated)

        :param ID user_id: user ID
        :param int? amt: amount of ratings by page (default: API default)
        :param int? page: page number (default: 1)
        :returns: {
            'has_next': bool,
            'next_page': int,
            'user_ratings': ratings array with fields
                ['item_id': ID, 'rating': float, 'timestamp': float]
            'warnings?': [str],
        }
        """
        user_id = self._userid2url(user_id)
        path = f'users/{user_id}/ratings/'
        params = {}
        if amt:
            params['amt'] = amt
        if page:
            params['page'] = page
        resp = self.api.get(path=path, params=params)
        resp['user_ratings'] = self._body2itemid(resp['user_ratings'])
        return resp

    @require_login
    def list_ratings(self, amt=None, cursor=None):
        """
        List the ratings of one database

        :param int? amt: amount to return (default: use the API default)
        :param str? cursor: Pagination cursor
        :returns: {
            'has_next': bool,
            'next_cursor': str,
            'ratings': array with fields
                ['item_id': ID, 'user_id': ID, 'rating': float, 'timestamp': float]
            'warnings?': [str],
        }
        """
        path = f'ratings-bulk/'
        params = {}
        if amt:
            params['amt'] = amt
        if cursor:
            params['cursor'] = cursor
        resp = self.api.get(path=path, params=params)
        resp['ratings'] = self._body2userid(self._body2itemid(resp['ratings']))
        return resp

    @require_login
    def delete_rating(self, user_id, item_id):
        """
        Delete a single rating for a given user.

        :param ID user_id: user ID
        :param ID item_id: item ID
        """
        user_id = self._userid2url(user_id)
        item_id = self._itemid2url(item_id)
        path = f'users/{user_id}/ratings/{item_id}'
        return self.api.delete(path=path)

    @require_login
    def delete_user_ratings(self, user_id):
        """
        Delete all ratings of a given user.

        :param ID user_id: user ID
        """
        user_id = self._userid2url(user_id)
        path = f'users/{user_id}/ratings/'
        return self.api.delete(path=path)

    # === User Interactions ===

    @require_login
    def create_interaction(self, user_id, item_id, interaction_type, timestamp=None):
        """
        This endpoint allows you to create a new interaction for a user and an item.
        An inferred rating will be created or updated for the tuple (user_id, item_id).
        The taste profile of the user will then be updated in real-time by the online machine learning algorithm.

        :param ID user_id: user ID
        :param ID item_id: item ID
        :param str interaction_type: Interaction type
        :param float? timestamp: rating timestamp (default: now)
        """
        user_id = self._userid2url(user_id)
        item_id = self._itemid2url(item_id)
        path = f'users/{user_id}/interactions/{item_id}/'
        data = {
            'interaction_type': interaction_type,
        }
        if timestamp is not None:
            data['timestamp'] = timestamp
        return self.api.post(path=path, data=data)

    @require_login
    def create_interactions_bulk(self, interactions, chunk_size=(1 << 14)):
        """
        Create or update large bulks of interactions for many users and many items.
        Inferred ratings will be created or updated for all tuples (user_id, item_id).

        :param array interactions: interactions array with fields:
            ['user_id': ID, 'item_id': ID, 'interaction_type': str, 'timestamp': float]
        :param int? chunk_size: split the requests in chunks of this size (default: 16_000)
        """
        path = f'interactions-bulk/'
        n_chunks = int(numpy.ceil(len(interactions) / chunk_size))
        sleep = chunk_size / 500
        for i in tqdm(range(n_chunks), disable=(True if n_chunks < 4 else None)):
            interactions_chunk = interactions[i*chunk_size:(i+1)*chunk_size]
            interactions_chunk = self._userid2body(self._itemid2body(interactions_chunk))
            data = {
                'interactions': interactions_chunk,
            }
            self.api.post(path=path, data=data, timeout=10)
            if n_chunks > 1:
                time.sleep(sleep)

        return

    @require_login
    def create_user_interactions_bulk(self, user_id, interactions):
        """
        Create a small bulk of interactions for a single user and many items.
        Inferred ratings will be created or updated for all tuples (user_id, item_id)

        :param ID user_id: user ID
        :param array interactions: interactions array with fields:
            ['item_id': ID, 'interaction_type': str, 'timestamp': float]
        :param int? chunk_size: split the requests in chunks of this size (default: 16_000)
        """
        path = f'users/{user_id}/interactions-bulk/'
        data = {
            'interactions': interactions,
        }
        return self.api.post(path=path, data=data, timeout=10)

    @require_login
    def list_interactions(
            self, amt=None, cursor=None,
            start_timestamp=None, end_timestamp=None
    ):
        """
        List the interactions of one database

        :param int? amt: amount to return (default: use the API default)
        :param str? cursor: Pagination cursor
        :param float? start_timestamp: When provided, only returns interactions
            timestamped after this value
        :param float? end_timestamp: When provided, only returns interactions
            timestamped before this value
        :returns: {
            'has_next': bool,
            'next_cursor': str,
            'interactions': array with fields
                ['item_id': ID, 'user_id': ID, 'interaction_type': str, 'timestamp': float]
            'warnings?': [str],
        }
        :raises: RequestError if provided timeframe is invalid
        """
        path = f'interactions-bulk/'
        params = {}
        if amt:
            params['amt'] = amt
        if cursor:
            params['cursor'] = cursor
        if start_timestamp:
            params['start_timestamp'] = start_timestamp
        if end_timestamp:
            params['end_timestamp'] = end_timestamp
        resp = self.api.get(path=path, params=params)
        resp['interactions'] = self._body2userid(self._body2itemid(resp['interactions']))
        return resp

    # === Session Interactions ===

    @require_login
    def create_session_interaction(self, session_id, item_id, interaction_type, timestamp=None):
        """
        This endpoint allows you to create a new interaction for an anonymous session and an item.

        :param ID session_id: anonymous session ID
        :param ID item_id: item ID
        :param str interaction_type: Interaction type
        :param float? timestamp: rating timestamp (default: now)
        :returns: {
            'warnings?': [str],
        }
        """
        session_id = self._sessionid2url(session_id)
        item_id = self._itemid2url(item_id)
        path = f'sessions/{session_id}/items/{item_id}/interactions/'
        data = {
            'interaction_type': interaction_type,
        }
        if timestamp is not None:
            data['timestamp'] = timestamp
        return self.api.post(path=path, data=data)

    @require_login
    def create_sessions_interactions_bulk(self, interactions, chunk_size=(1 << 14)):
        """
        Create or update large bulks of interactions for many anonymous session and many items.

        :param array interactions: interactions array with fields:
            ['session_id': ID, 'item_id': ID, 'interaction_type': str, 'timestamp': float]
        :param int? chunk_size: split the requests in chunks of this size (default: 16_000)
        :returns: {
            'warnings?': [str],
        }
        """
        path = f'sessions/interactions-bulk/'
        n_chunks = int(numpy.ceil(len(interactions) / chunk_size))
        sleep = chunk_size / 500
        warnings = []
        for i in tqdm(range(n_chunks), disable=(True if n_chunks < 4 else None)):
            interactions_chunk = interactions[i * chunk_size:(i + 1) * chunk_size]
            interactions_chunk = self._sessionid2body(self._itemid2body(interactions_chunk))
            data = {
                'interactions': interactions_chunk,
            }
            resp = self.api.post(path=path, data=data, timeout=10)
            warnings.extend(resp.get('warnings', []))
            if n_chunks > 1:
                time.sleep(sleep)
        return {'warnings': warnings} if warnings else {}

    @require_login
    def create_session_interactions_bulk(self, session_id, interactions):
        """
        Create a small bulk of interactions for a single anonymous session and many items.

        :param ID session_id: anonymous session ID
        :param array interactions: interactions array with fields:
            ['item_id': ID, 'interaction_type': str, 'timestamp': float]
        :param int? chunk_size: split the requests in chunks of this size (default: 16_000)
        :returns: {
            'warnings?': [str],
        }
        """
        session_id = self._sessionid2url(session_id)
        path = f'sessions/{session_id}/interactions-bulk/'
        data = {
            'interactions': self._itemid2body(interactions),
        }
        return self.api.post(path=path, data=data, timeout=10)

    @require_login
    def list_sessions_interactions(
            self, amt=None, cursor=None,
            start_timestamp=None, end_timestamp=None
    ):
        """
        List the anonymous sessions interactions of one database

        :param int? amt: amount to return (default: use the API default)
        :param str? cursor: Pagination cursor
        :param float? start_timestamp: When provided, only returns interactions
            timestamped after this value
        :param float? end_timestamp: When provided, only returns interactions
            timestamped before this value
        :returns: {
            'has_next': bool,
            'next_cursor': str,
            'interactions': array with fields
                ['item_id': ID, 'session_id': ID, 'interaction_type': str, 'timestamp': float]
            'warnings?': [str],
        }
        :raises: RequestError if provided timeframe is invalid
        """
        path = f'sessions/interactions-bulk/'
        params = {}
        if amt:
            params['amt'] = amt
        if cursor:
            params['cursor'] = cursor
        if start_timestamp:
            params['start_timestamp'] = start_timestamp
        if end_timestamp:
            params['end_timestamp'] = end_timestamp
        resp = self.api.get(path=path, params=params)
        resp['interactions'] = self._body2sessionid(self._body2itemid(resp['interactions']))
        return resp

    # === Scheduled Background Tasks ===

    @require_login
    def trigger_background_task(self, task_name, payload=None):
        """
        Trigger background task such as retraining of ML models.
        You should not have to call this endpoint yourself, as this is done automatically.

        :param str task_name: for instance ``'ml_model_retrain'``
        :param dict? payload: optional task payload
        :returns: {
            'task_id': str,
            'warnings?': [str],
        }
        :raises: DuplicatedError with error name 'TASK_ALREADY_RUNNING'
            if this task is already running
        """
        path = f'tasks/{self.escape_url(task_name)}/'
        data = {}
        if payload:
            data['payload'] = payload
        return self.api.post(path=path, data=data)

    @require_login
    def get_background_tasks(self, task_name, page=None, amt=None):
        """
        List currently running background tasks such as ML models training.

        :param str task_name: names allowed ``ml_model_retrain``,
        ``item_popularity_score_recalibrate``, ``ml_new_users``, ``ml_new_items``
        :param int? page: page number (default: 1)
        :param int? amt: amount of tasks by page (default: 50)
        :returns: {
            'tasks': [{
                'task_id': str,
                'name': str,
                'start_time': int,
                'status': str,
                'progress?': str,
            }],
            'warnings?': [str],
        }
        """
        params = {}
        if page is not None:
            params['page'] = page
        if amt is not None:
            params['amt'] = amt
        path = f'tasks/{self.escape_url(task_name)}/recents/'
        return self.api.get(path=path, params=params)

    def wait_until_ready(self, timeout=600, sleep=1):
        """
        Wait until the current database status is ready, meaning at least one model has been trained

        :param int? timeout: maximum time to wait, raise RuntimeError if exceeded (default: 10min)
        :param int? sleep: time to wait between polling (default: 1s)
        """
        assert sleep > 0.1
        resp = None
        time_start = time.time()
        while time.time() - time_start < timeout:
            time.sleep(sleep)
            resp = self.status()
            if resp['status'] == 'ready':
                return
        raise RuntimeError(f'API not ready before {timeout}s. Last response: {resp}')

    @require_login
    def trigger_and_wait_background_task(self, task_name, timeout=600, lock_wait_timeout=None,
                                         sleep=1, verbose=None):
        """
        Trigger background task such as retraining of ML models.
        You don't necessarily have to call this endpoint yourself,
        model training is also triggered automatically.
        By default this waits for an already running task before triggering the new one

        :param str task_name: for instance ``'ml_model_retrain'``
        :param int? timeout: maximum time to wait after the new task is triggered (default: 10min)
        :param int? lock_wait_timeout: if another task is already running, maximum time to wait
            for it to finish before triggering the new task (default: ``timeout``)
        :param int? sleep: time to wait between polling (default: 1s)
        :returns: {
            'task_id': str,
            'warnings?': [str],
        }
        :raises: RuntimeError if either ``timeout`` or ``lock_wait_timeout`` is reached
        """
        assert sleep > 0.1
        if lock_wait_timeout is None:
            lock_wait_timeout = timeout
        if verbose is None:
            verbose = sys.stdout.isatty()
        # wait for already running task (if any)
        if lock_wait_timeout > 0:
            msg = 'waiting for already running...' if verbose else None
            self.wait_for_background_task(
                task_name, lock_wait_timeout, sleep, msg=msg, wait_if_no_task=False,
                filtr=lambda t: t['status'] == 'RUNNING')
        # trigger
        try:
            task_id = self.trigger_background_task(task_name)['task_id']
        except DuplicatedError as exc:
            if getattr(exc, 'data', {})['name'] != 'TASK_ALREADY_RUNNING':
                raise
            # edge case: something else triggered the same task at the same time
            tasks = self.get_background_tasks(task_name)['tasks']
            task_id = next(t['task_id'] for t in tasks if t['status'] != 'COMPLETED')
        # wait for new task
        msg = 'waiting...' if verbose else None
        self.wait_for_background_task(
            task_name, timeout, sleep, msg=msg, filtr=lambda t: t['task_id'] == task_id)

    def wait_for_background_task(self, task_name, timeout=600, sleep=1, msg=None, filtr=None,
                                 wait_if_no_task=True, func_list_tasks=None):
        """
        Wait for a certain background task. Optionally specified with ``filtr`` function

        :param str task_name: for instance ``'ml_model_retrain'``
        :param int? timeout: maximum time to wait after the new task is triggered (default: 10min)
        :param int? sleep: time to wait between polling (default: 1s)
        :param str? msg: either ``None`` to disable print, or message prefix (default: None)
        :param func? filtr: filter function(task: bool)
        :param bool? wait_if_no_task: wait (instead of return) if there is no task satisfying filter
        :param func? func_list_tasks: function to list background tasks (default: self.get_background_tasks)
        :returns: bool; True is a task satisfying filters successfully ran, False otherwise
        :raises: RuntimeError if ``timeout`` is reached or if the task failed
        """
        spinner = '|/-\\'
        task = None
        time_start = time.time()
        time_waited = 0
        i = 0
        func_list_tasks = func_list_tasks or self.get_background_tasks
        while time_waited < timeout:
            time.sleep(sleep)
            time_waited = time.time() - time_start
            print_time = f'{int(time_waited) // 60:d}m{int(time_waited) % 60:02d}s'
            tasks = func_list_tasks(task_name)['tasks']
            try:
                task = max(filter(filtr, tasks), key=lambda t: t['start_time'])
            except ValueError:
                if wait_if_no_task:
                    continue
                else:
                    return (task is not None)
            progress = task.get('progress', '')
            if task['status'] == 'COMPLETED':
                if msg is not None:
                    print(f'\r{msg} {print_time} done   {progress:80s}')
                return True
            elif task['status'] == 'FAILED':
                raise RuntimeError(f'task {task_name} failed with: {progress}')
            if msg is not None:
                print(f'\r{msg} {print_time} {spinner[i%len(spinner)]} {progress:80s}', end='')
                sys.stdout.flush()
            i += 1
        raise RuntimeError(f'task {task_name} not done before {timeout}s. Last response: {task}')

    # === Pre-trained Machine Learning ===

    @require_login
    def list_trained_models(self):
        """
        Get all trained models
        :return: trained models
        {
            'trained_models': [
                {
                    'id': str,
                    'trained_timestamp': str
                    'algorithm': str,
                }
            ],
            'warnings?': [str],
        }
        """
        path = f'pretrained-ml/trained-models/'
        return self.api.get(path=path)

    # === Scenarios ===

    @require_login
    def get_scenario(self, reco_type, name):
        """
        Get a scenario
        :param str reco_type
        :param str name:
        :raise: NotFoundError if not found
        :return:
        {
            'name': str,
            'reco_type': str
            'scenario': {
                'scenario_type': 'case',
                'case': {
                    'filters?': [dict],
                    'reranking?': [dict],
                    'exclude_rated_items?': bool,
                    'algorithms?': str,
                    'candidates_preselection?': dict,
                }
            },
            'warnings?': [str],
        }
        or
        {
            'name': str,
            'reco_type': str
            'scenario': {
                'scenario_type': 'ab_test',
                'ab_test': {
                    'id': str,
                    'scenario_a': str,
                    'scenario_b': str
                }
            },
            'warnings?': [str],
        }
        or
        {
            'name': str,
            'reco_type': str
            'scenario': {
                'scenario_type': 'condition',
                'condition': {
                    'condition_type': str,
                    'if': dict,
                    'then': str,
                    'else': str
                }
            },
            'warnings?': [str],
        }
        """
        path = f'scenarios/{reco_type}/{name}/'
        return self.api.get(path=path)

    @require_login
    def list_scenarios(self):
        """
        Get all scenarios
        :return: scenarios
        {
            'scenarios': [
                {
                    'name': str,
                    'reco_type': str
                    'scenario': {
                        'scenario_type': str,
                        'case?ab_test?condition?': dict
                    }
                }
            ],
            'warnings?': [str],
        }
        """
        path = f'scenarios/'
        return self.api.get(path=path)

    @require_login
    def create_or_replace_scenario(self, reco_type, name, scenario):
        """
        Create a new scenario or replace an existing one.
        Only `case` scenarios can be replaced, to avoid circular referrencing.

        A scenario should take the following form, with only one of the keys
        among case|condition|ab_test, corresponding to the scenario type:

        :param str reco_type: accepted values
            "item_to_items", "profile_to_items", "session_to_items"
        :param str name: name of the scenario
        :param dict scenario: a ``?`` in a key means the key is optional
            {
                'scenario_type': str,   # 'case', 'ab_test' or 'condition'
                'case?: {
                    'filters?': [dict],
                    'reranking?': [dict],
                    'exclude_rated_items?': bool,
                    'algorithms?': str,
                    'candidates_preselection?': dict,
                },
                'ab_test?': {
                    'id': str,
                    'scenario_a': str,
                    'scenario_b': str
                },
                'condition': {
                    'condition_type': str,
                    'if': dict,  # FIXME
                    'then': str,   # scenario name
                    'else': str    # scenario name
                }
            }
        :returns: {'warnings?': [str]}
        :raise: RequestError if some business rule is invalid
        :raise: NotFoundError if an AB-test or a then/else scenario is not found

        Example
        _______
        api.create_or_replace_scenario(
            reco_type='profile_to_items',
            name='my_case_scenario',
            scenario = {
                'scenario_type': 'case',
                'case': {
                    'filters': [
                            {'property_name': 'tags', 'op': 'EQ', 'value': 'pi'},
                            {'property_name': 'price', 'op': 'GEQ', 'value': 3.14}
                    ],
                    'reranking': [
                            {'property_name"': 'director', 'op': 'diversity', 'weight': 0.8}
                    ],
                    'algorithms': 'algo1|algo2',
                    'candidates_preselection': {
                        'method': 'rated_items_nearest_neighbor',
                        'params': {'nearest_neighbor_algorithms': ['algo3']}
                    },
                    # only for reco types "profile_to_items", "session_to_items":
                    'exclude_rated_items': True,
                }
            }
        )

        Example
        _______
        api.create_or_replace_scenario(
            reco_type='profile_to_items',
            name='my_ab_test_scenario',
            scenario = {
                'scenario_type': 'ab_test',
                'ab_test': {
                    'id': 'my_ab_test_id',
                    'scenario_a': 'scenar1',
                    'scenario_b': 'scenar2'
                }
            }
        )

        Example
        _______
        api.create_or_replace_scenario(
            reco_type='profile_to_items',
            name='my_condition_scenario',
            scenario = {
                'scenario_type': 'condition',
                'condition': {
                    'condition_type': 'user_function',
                    'if': {
                        'function_name': 'n_ratings',
                        'op': 'lt',
                        'value': 10
                    },
                    'then': 'scenar1',
                    'else': 'scenar2'
                }
            }
        )
        """
        path = f'scenarios/{reco_type}/{name}/'
        data = scenario
        return self.api.put(path=path, data=data)

    @require_login
    def delete_scenario(self, reco_type, name):
        """
        Delete scenario with provided name.
        :param str reco_type:
        :param str name:
        :raise: NotFoundError if not found
        """
        path = f'scenarios/{reco_type}/{name}/'
        self.api.delete(path=path)

    # === Default Scenario ===

    @require_login
    def get_default_scenario(self, reco_type):
        """
        Get default scenario
        :param str reco_type:
        :raise: NotFoundError if no default scenario
        :return:
        {
            'name': str,
            'warnings?': [str],
        }
        """
        path = f'scenarios-default/{reco_type}/'
        return self.api.get(path=path)

    @require_login
    def set_default_scenario(self, reco_type, name):
        """
        Set scenario as default for a reco type
        """
        path = f'scenarios-default/{reco_type}/'
        data = {'name': name}
        self.api.patch(path=path, data=data)

    @require_login
    def unset_default_scenario(self, reco_type):
        """
        Unset default scenario for a reco type
        """
        path = f'scenarios-default/{reco_type}/'
        self.api.delete(path=path)

    # === Utils ===

    def clear_jwt_token(self):
        return self.api.clear_jwt_token()

    @property
    def jwt_token(self):
        return self.api.jwt_token

    @property
    def last_trace(self):
        return self.api.last_trace

    def set_jwt_token(self, jwt_token):
        self.api.set_jwt_token(jwt_token)

    def escape_url(self, param):
        return quote(param, safe='')

    def _userid2url(self, user_id):
        """ base64 encode if needed """
        return self._id2url(user_id, 'user')

    def _itemid2url(self, item_id):
        """ base64 encode if needed """
        return self._id2url(item_id, 'item')

    def _sessionid2url(self, item_id):
        """ base64 encode if needed """
        return self._id2url(item_id, 'session')

    def _id2url(self, data, field):
        """ base64 encode if needed """
        assert self._database is not None, f'You need to login to a database first'
        if self._database[f'{field}_id_type'].startswith('bytes'):
            return self._b64_encode(data)
        if isinstance(data, bytes):
            return data.decode('ascii')
        return data

    def _userid2body(self, data):
        return self._base_field_id(data, 'user', self._id2body)

    def _itemid2body(self, data):
        return self._base_field_id(data, 'item', self._id2body)

    def _sessionid2body(self, data):
        return self._base_field_id(data, 'session', self._id2body)

    def _body2userid(self, data):
        return self._base_field_id(data, 'user', self._body2id)

    def _body2itemid(self, data):
        return self._base_field_id(data, 'item', self._body2id)

    def _body2sessionid(self, data):
        return self._base_field_id(data, 'session', self._body2id)

    def _base_field_id(self, data, field, cast_func):
        if not self.b64_encode_bytes:
            return data
        d_type = self._database[f'{field}_id_type']
        if not d_type.startswith(('bytes', 'uuid', 'hex', 'urlsafe')):
            return data
        if isinstance(data, numpy.ndarray):
            if data.dtype.names:
                data = [dict(zip(data.dtype.names, row)) for row in data.tolist()]
            else:
                data = data.tolist()
        if not isinstance(data, list):
            return cast_func(data, d_type)
        if all(isinstance(d, dict) for d in data):
            return [{**row, f'{field}_id': cast_func(row[f'{field}_id'], d_type)} for row in data]
        return [cast_func(row, d_type) for row in data]

    def _id2body(self, data, d_type):
        if d_type.startswith(('uuid', 'hex', 'urlsafe')):
            return data.decode('ascii') if isinstance(data, bytes) else data
        else:  # Bytes
            return self._b64_encode(data)

    def _body2id(self, data, d_type):
        if d_type.startswith(('uuid', 'hex', 'urlsafe')):
            return data  # leave as str, do not encode
        else:  # Bytes
            return self._b64_decode(data.encode('ascii'))

    def _b64_encode(self, data):
        return base64.urlsafe_b64encode(data).replace(b'=', b'').decode('ascii')

    def _b64_decode(self, data):
        n_pad = (4 - len(data) % 4) % 4
        if n_pad <= 2:
            data = data + b'=' * n_pad
        elif n_pad == 3:
            raise TypeError()  # TODO
        try:
            return base64.b64decode(data, b'-_')
        except BinasciiError:
            raise TypeError()

    def _get_latest_task_progress_message(self, task_name,
                                          default=None, default_running=None, default_failed=None):
        tasks = self.get_background_tasks(task_name)['tasks']
        if not tasks:
            return default
        latest_task = max(tasks, key=lambda t: t['start_time'])
        if latest_task['status'] == 'RUNNING':
            return latest_task.get('progress', default_running)
        if latest_task['status'] == 'FAILED':
            raise ServerError({'error': latest_task.get('error', default_failed)})
        return latest_task['status']

    @staticmethod
    def _clean_filters_to_urlsafe(filters):
        """
        Cleans item-property filters for GET recommendation endpoints
        Filter format: '<PROP_NAME>:<OPERATOR>:<OPTIONAL_VALUE>'

        :param filters: list-of-(dict-or-str)
        :return: list-of-str

        Example:

        >>> CrossingMindsApiClient._clean_filters_to_urlsafe([
        ...     {'property_name': 'test', 'op': 'gt', 'value': 1.0},
        ...     {'property_name': 'test', 'op': 'notempty'}
        ... ])
        ['test:gt:1.0', 'test:notempty']
        """
        _filters = []
        for f in filters:
            if isinstance(f, str):
                _f = f
            elif isinstance(f, dict):
                _f = f"{f['property_name']}:{f['op']}"
                if 'value' in f:
                    v = f['value']
                    if isinstance(v, dict):
                        raise TypeError('Advanced filtering features are reserved to scenarios')
                    _f += f':{v}'
            else:
                raise TypeError(
                    f'only strings and dict are allowed but one is a {type(f)}')
            _filters.append(_f)
        return _filters

    @staticmethod
    def _clean_reranking_to_urlsafe(reranking):
        """
        Cleans rerankings for GET recommendation endpoints
        Format: ['<PROP_NAME>:<OPERATOR>:<OPTIONAL_WEIGHT>:<OPTION_TYPE>:<OPTION_VALUE>']

        :param reranking: list-of-(dict-or-str)
        :return: list-of-str.

        Example:

        >>> CrossingMindsApiClient._clean_reranking_to_urlsafe([
        ...     {'property_name': 'test', 'op': 'diversity', 'weight': 1, 'options': {'a': 2, 'b': 3}},
        ...     {'property_name': 'test', 'op': 'diversity', 'options': {'a': 1}}
        ... ])
        ['test:diversity:1:a:2:b:3', 'test:diversity::a:1']
        """
        _reranking = []
        for r in reranking:
            if isinstance(r, str):
                _r = r
            elif isinstance(r, dict):
                _r = f"{r['property_name']}:{r['op']}:"
                weight = r.get('weight')
                if weight is not None:
                    _r += f"{weight}"
                options = r.get('options')
                if options is not None:
                    s = []
                    for k, v in options.items():
                        assert not isinstance(v, dict)
                        s.append(f'{k}:{v}')
                    _r += ':' + ':'.join(s)
            else:
                raise TypeError(
                    f'only strings and dict are allowed but one is a {type(r)}')
            _reranking.append(_r)
        return _reranking



import numpy


def structured_cast(array, dtype):
    """
    if array is a structured array, align fields by names and cast to dtype
    if any field with name '__pad__' is present in `dtype` but not in `array`,
    then output will include padding zeros
    :param array array:
    :param dtype-descr dtype:
    :returns: array
    """
    dtype = numpy.dtype(dtype)
    if array.dtype == dtype:
        return array
    if not dtype.names:
        return array.astype(dtype)
    if not('__pad__' in dtype.names and '__pad__' not in array.dtype.names):
        return array[list(dtype.names)].astype(dtype)

    new_array = numpy.empty_like(array, dtype=dtype)
    for field_name, (field_dtype, _) in dtype.fields.items():
        if field_name == '__pad__':
            new_array[field_name] = numpy.zeros(1, field_dtype)[0]
        else:
            new_array[field_name] = array[field_name].astype(field_dtype)
    return new_array



"""
Aggregate tools
===============
"""
import sys

import numpy

from .._lib.hashmap import factorize
from ..compat import tqdm
from ..ds.scaling import linearscaling
from .arrays import first, lexsort_uint32_pair, to_structured


def igroupby(ids, values, n=None, logging_prefix=None, assume_sorted=False, stable=False,
             find_next_hint=512):
    """
    Efficiently converts two arrays representing a relation
    (the ``ids`` and the associated ``values``) to an iterable ``(id, values_associated)``.

    The ``values`` are grouped by ``ids`` and a sequence of tuples is generated.

    The ``i`` th tuple generated is ``(id_i, values[ids == id_i])``,
    ``id_i`` being the ``i`` th element of the ``ids`` array, once sorted in ascending order.

    :param array ids: ``(>=n,) dtype array``
    :param array values: ``(>=n, *shape) uint32 array``
    :param int? n: length of array to consider,
     applying igroupby to ``(ids[:n], values[:n])``. Uses full array when not set.
    :param string? logging_prefix: prefix to include while logging progress.
        ``(default:`` Does not log``)``.
    :param bool? assume_sorted: whether ids is sorted. ``(default: False)``
    :param bool? stable: when True, ensures that the relative order of values is preserved
        (will slightly impact performance)
        When False, nothing guarantees that the values in ``(id_i, values[ids == id_i])``
        will be aligned on the input values (see example below)
        The stable option is only used when `assume_sorted` is False (general case)
    :param int? find_next_hint: hint for find_next_lookup. ``(default: 512)``
    :generates: tuple(id:int, values_associated:``(m, *shape) array slice``)

    Example
    _______
    >>> ids     = numpy.array([1, 1, 1, 1, 1, 0, 0, 0, 0, 3, 3, 3, 3, 3, 3, 3, 3,  3,  3,  3])
    >>> values  = numpy.array([0, 1, 2, 3, 4, 0, 2, 4, 6, 0, 4, 6, 5, 6, 7, 8, 9, 10, 11, 12])
    >>> gen = igroupby(ids, values)
    >>> next(gen)
    (0, array([0, 2, 4, 6]))

    >>> next(gen)
    (1, array([0, 1, 2, 3, 4]))

    >>> next(gen)
    (3, array([10,  9,  8,  7,  0,  5,  6,  4, 11,  6, 12]))
    # note how the values are not in the same order as they were in the input

    with option `stable`:
    >>> gen = igroupby(ids, values, stable=True)
    >>> next(gen)
    >>> next(gen)
    >>> next(gen)
    (3, array([0, 4, 6, 5, 6, 7, 8, 9, 10, 11, 12]))
    # this time, the values are in the same order as they were in the input

    Example with strings as ids:

    >>> ids = numpy.array(["alpha", "alpha", "beta", "omega", "alpha", "gamma", "beta"])
    >>> values = numpy.array([1, 2, 10, 100, 3, 1000, 20])
    >>> gen = igroupby(ids, values)
    >>> next(gen)
    ('alpha', array([1, 2, 3]))
    >>> next(gen)
    ('beta', array([10, 20]))
    >>> next(gen)
    ('gamma', array([1000]))
    >>> next(gen)
    ('omega', array([100]))
    """
    # convert to numpy arrays
    ids = numpy.asarray(ids)
    values = numpy.asarray(values)
    # check input shape
    assert len(ids.shape) == 1
    if n is None:
        n = ids.shape[0]
    assert ids.shape[0] >= n and values.shape[0] >= n, values.shape
    # sort if needed
    if not assume_sorted:
        ids = ids[:n]
        values = values[:n]
        if stable:
            # slower, but ensures relative order of values will be preserved
            asort = numpy.argsort(ids, kind='stable')
        else:
            asort = numpy.argsort(ids)
        ids = ids[asort]
        values = values[asort]
    # init
    start_block = 0
    find_next_lookup = find_next_hint
    # search next change block by block
    disable = logging_prefix is None
    with tqdm(total=n, desc=logging_prefix, disable=disable, file=sys.stdout) as pbar:
        while start_block < n:
            # find all items having id by block boundaries
            current_id = ids[start_block]
            try:
                end_block = first(ids,
                                  lambda x: x != current_id,
                                  offset=start_block,
                                  batch_size=find_next_lookup)
                find_next_lookup = max(
                    find_next_hint, 2 * (end_block - start_block))
            except StopIteration:
                end_block = n
            current_id_values = values[start_block:end_block]
            assert (ids[start_block:end_block] == current_id).all()
            pbar.update(end_block - start_block)
            start_block = end_block
            yield current_id, current_id_values


def ufunc_group_by_idx(idx, values, ufunc, init, minlength=None):
    """
    Abstract wrapper to compute ufunc grouped by values in array ``idx``.

    Return an array containing the results of ``ufunc`` applied to ``values``
    grouped by the indexes in array ``idx``.
    (See available ufuncs `here <https://docs.scipy.org/doc/numpy/reference/ufuncs.html>`_).

    Warning: the ``init`` parameter is not a filling value for missing indexes.
    If index ``i`` is missing, then ``out[i] = init``
    but this value also serves as the initialization of ``ufunc`` on all the groups of ``values``.

    For example, if ``ufunc`` is ``numpy.add`` and ``init = -1`` then for each index,
    the sum of the corresponding values will be decreased by one.

    :param array idx: ``(n,) int array``
    :param array values: ``(n,) dtype array``
    :param numpy.ufunc ufunc: universal function applied to the groups of ``values``
    :param dtype init: initialization value
    :param int? minlength: ``(default: idx.max() + 1)``
    :returns: (min-length,) dtype array, such that ``out[i] = ufunc(values[idx==i])``

    Example
    _______
    >>> idx  = numpy.array([0, 0, 0, 0, 0, 1, 1, 1, 1, 3, 3, 3])
    >>> values   = numpy.array([0, 1, 2, 3, 4, 0, 2, 4, 6, 0, 4, 6])
    >>> ufunc_group_by_idx(idx, values, numpy.maximum, -1)
    array([ 4, 6,  -1,  6])
    >>> ufunc_group_by_idx(idx, values, numpy.add, -1)
    array([ 9, 11, -1,  9])
    >>> ufunc_group_by_idx(idx, values, numpy.add, 0)
    array([ 10, 12, -0,  10])
    """
    length = max(idx.max() + 1 if idx.size else 0, minlength or 0)
    out = numpy.full(length, init)
    ufunc.at(out, idx, values)
    return out


def min_by_idx(idx, values, minlength=None, fill=None):
    """
    Given array of indexes ``idx`` and array ``values``,
    outputs the max value by idx, aligned on ``arange(idx.max() + 1)``.
    See also ``argmin_by_idx`` and ``value_at_argmin_by_idx``.

    :param array idx: (n,) int array
    :param array values: (n,) float array
    :param int? minlength: (default: idx.max() + 1)
    :param float? fill: filling value for missing idx (default: +inf)
    :returns: (min-length,) float array, such that out[i] = min(values[idx==i])

    Example
    _______
    >>> idx  = numpy.array([0, 0, 0, 0, 0, 1, 1, 1, 1, 3, 3, 3])
    >>> values   = numpy.array([1, 1, 2, 3, 4, 0, 2, 4, 6, 0, 4, 6])
    >>> min_by_idx(idx, values, fill=100)
    array([  1,   0, 100,   0])
    >>> min_by_idx(idx, values)
    array([1, 0, 9223372036854775807, 0])
    """
    assert idx.dtype.kind == 'u' or (idx.dtype.kind == 'i' and (idx >= 0).all()), (
        'Can only use get_xx_by_idx with integer idx, where (idx >= 0).all()')
    if fill is None:
        fill = numpy.inf if values.dtype.kind == 'f' else numpy.iinfo(values.dtype).max
    else:
        assert fill >= values.max()
    return ufunc_group_by_idx(idx, values, numpy.minimum, fill, minlength=minlength)


def max_by_idx(idx, values, minlength=None, fill=None):
    """
    Given array of indexes ``idx`` and array ``values``,
    outputs the max value by idx, aligned on ``arange(idx.max() + 1)``.
    See also ``argmax_by_idx`` and ``value_at_argmax_by_idx``.

    :param array idx: (n,) int array
    :param array values: (n,) float array
    :param int? minlength: (default: idx.max() + 1)
    :param float? fill: filling value for missing idx (default: -inf)
    :returns: (min-length,) float array, such that out[i] = max(values[idx==i])

    Example
    _______
    >>> idx  = numpy.array([0, 0, 0, 0, 0, 1, 1, 1, 1, 3, 3, 3])
    >>> values   = numpy.array([0, 1, 2, 3, 4, 0, 2, 4, 6, 0, 4, 6])
    >>> max_by_idx(idx, values, fill=-1)
    array([ 4, 6,  -1,  6])
    >>> max_by_idx(idx, values, minlength=10, fill=-1)
    array([ 4,  6, -1,  6, -1, -1, -1, -1, -1, -1])
    >>> max_by_idx(idx, values)
    array([ 4, 6, -9223372036854775808, 6])
    """
    assert idx.dtype.kind == 'u' or (idx.dtype.kind == 'i' and (idx >= 0).all()), (
        'Can only use get_xx_by_idx with integer idx, where all idx >= 0')
    if fill is None:
        fill = - numpy.inf if values.dtype.kind == 'f' else numpy.iinfo(values.dtype).min
    else:
        assert fill <= values.min()
    return ufunc_group_by_idx(idx, values, numpy.maximum, fill, minlength=minlength)


def argmin_by_idx(idx, values, minlength=None, fill=None):
    """
    Given array of indexes ``idx`` and array ``values``,
    outputs the argmin of the values by idx,
    aligned on ``arange(idx.max() + 1)``.
    See also ``min_by_idx`` and ``value_at_argmin_by_idx``.

    :param array idx: (n,) int array
    :param array values: (n,) float array
    :param int? minlength: (default: idx.max() + 1)
    :param float? fill: filling value for missing idx (default: -1)
    :returns: (min-length,) int32 array, such that
              out[i] = argmin_{idx}(values[idx] : idx[idx] == i)

    Example
    _______
    >>> idx  = numpy.array([0, 0, 0, 0, 0, 1, 1, 1, 1, 3, 3, 3])
    >>> values   = numpy.array([0, 1, 2, 3, 4, 0, 2, 4, 6, 0, 4, 6])
    >>> argmin_by_idx(idx, values, fill=-1)
    array([ 0,  5, -1,  9])
    >>> argmin_by_idx(idx, values, minlength=10, fill=-1)
    array([ 0,  5, -1,  9, -1, -1, -1, -1, -1, -1])
    """
    assert idx.dtype.kind == 'u' or (idx.dtype.kind == 'i' and (idx >= 0).all()), (
        'Can only use get_xx_by_idx with integer idx, where all idx >= 0')
    if fill is None:
        fill = -1
    min_values_by_idx = min_by_idx(idx, values, minlength)  # (n-idx,)
    is_min = values == min_values_by_idx[idx]
    out = numpy.full(min_values_by_idx.size, fill)
    out[idx[is_min]] = numpy.where(is_min)[0]
    return out


# TODO: improve test
def value_at_argmin_by_idx(idx, sorting_values, fill, output_values=None, minlength=None):
    """
    Wrapper around argmin_by_idx and get_value_by_idx.
    Allows to use a different value for the output and for detecting the minimum
    Allows to set a specific fill value that is not compared with the sorting_values

    :param array idx: (n,) uint array with values < max_idx
    :param array values: (n,) array
    :param fill: filling value for output[i] if there is no idx == i
    :param array? output_values: (n,) dtype array
        Useful if you want to select the min based on one array,
        and get the value on another array
    :param int? minlength: minimum shape for the output array.
    :returns array: (max_idx+1,), dtype array such that
        out[i] = min(values[idx==i])

    Example
    _______
    >>> idx  = numpy.array([0, 0, 0, 0, 0, 1, 1, 1, 1, 3, 3, 3])
    >>> values   = numpy.array([0, 1, 2, 3, 4, 0, 2, 4, 6, 0, 4, 6])
    >>> value_at_argmin_by_idx(idx, values, fill=-1)
    array([ 0,  0, -1,  0])
    >>> value_at_argmin_by_idx(idx, values, minlength=10, fill=-1)
    array([ 0,  0, -1,  0, -1, -1, -1, -1, -1, -1])
    """
    assert idx.dtype.kind == 'u' or (idx.dtype.kind == 'i' and (idx >= 0).all()), (
        'Can only use get_xx_by_idx with integer idx, where all idx >= 0')
    length = max(idx.max() + 1, minlength or 0)
    if output_values is None:
        output_values = sorting_values
    out = numpy.full(length, fill, dtype=output_values.dtype)
    argmin = argmin_by_idx(idx,
                           sorting_values,
                           minlength=minlength)
    mask = (argmin != -1)
    out[:mask.size][mask] = output_values[argmin[mask]]
    return out


def argmax_by_idx(idx, values, minlength=None, fill=None):
    """
    Given array of indexes ``idx`` and array ``values``,
    outputs the argmax of the values by idx,
    aligned on ``arange(idx.max() + 1)``.
    See also ``max_by_idx`` and ``value_at_argmax_by_idx``.

    :param array idx: (n,) int array
    :param array values: (n,) float array
    :param int? minlength: (default: idx.max() + 1)
    :param float? fill: filling value for missing idx (default: -1)
    :returns: (min-length,) int32 array, such that
              out[i] = argmax_{idx}(values[idx] : idx[idx] == i)

    Example
    _______
    >>> idx  = numpy.array([0, 0, 0, 0, 0, 1, 1, 1, 1, 3, 3, 3])
    >>> values   = numpy.array([0, 1, 2, 3, 4, 0, 2, 4, 6, 0, 4, 6])
    >>> argmax_by_idx(idx, values, fill=-1)
    array([ 4,  8, -1, 11])
    >>> argmax_by_idx(idx, values, minlength=10, fill=-1)
    array([ 4,  8, -1, 11, -1, -1, -1, -1, -1, -1])
    """
    assert idx.dtype.kind == 'u' or (idx.dtype.kind == 'i' and (idx >= 0).all()), (
        'Can only use get_xx_by_idx with integer idx, where all idx >= 0')
    if fill is None:
        fill = -1
    max_values_by_idx = max_by_idx(idx, values, minlength)  # (n-idx,)
    is_max = values == max_values_by_idx[idx]
    out = numpy.full(max_values_by_idx.size, fill)
    out[idx[is_max]] = numpy.where(is_max)[0]
    return out


# TODO: improve test
def value_at_argmax_by_idx(idx, sorting_values, fill, output_values=None, minlength=None):
    """
    Wrapper around ``argmax_by_idx`` and ``get_value_by_id``.
    Allows to use a different value for the output and for detecting the minimum
    Allows to set a specific fill value that is not compared with the sorting_values

    :param array idx: (n,) uint array with values < max_idx
    :param array values: (n,) array
    :param fill: filling value for output[i] if there is no idx == i
    :param array? output_values: (n,) dtype array
        Useful if you want to select the min based on one array,
        and get the value on another array
    :param int? minlength: minimum shape for the output array.
    :returns array: (max_idx+1,), dtype array such that
        out[i] = max(values[idx==i])

    Example
    _______
    >>> idx  = numpy.array([0, 0, 0, 0, 0, 1, 1, 1, 1, 3, 3, 3])
    >>> values   = numpy.array([0, 1, 2, 3, 4, 0, 2, 4, 6, 0, 4, 6])
    >>> value_at_argmax_by_idx(idx, values, fill=-1)
    array([ 4,  6, -1,  6])
    >>> value_at_argmax_by_idx(idx, values, minlength=10, fill=-1)
    array([ 4,  6, -1,  6, -1, -1, -1, -1, -1, -1])
    """
    assert idx.dtype.kind == 'u' or (idx.dtype.kind == 'i' and (idx >= 0).all()), (
        'Can only use get_xx_by_idx with integer idx, where all idx >= 0')
    length = max(idx.max() + 1, minlength or 0)
    if output_values is None:
        output_values = sorting_values
    out = numpy.full(length, fill, dtype=output_values.dtype)
    argmax = argmax_by_idx(idx,
                           sorting_values,
                           minlength=minlength)
    mask = (argmax != -1)
    out[:mask.size][mask] = output_values[argmax[mask]]
    return out


def connect_adjacents_in_groups(group_ids, values, max_gap):
    """
    For each group_id in ``group_ids``, connect values that are closer than ``max_gap`` together.

    Return an array mapping the values to the indexes of
    the newly formed connected components they belong to.

    Two values that don't have the same input group_id can's be connected in the same
    connected component.

    ``connect_adjacents_in_groups`` is faster when an array of indexes is provided as ``group_ids``,
    but also accepts other types of ids.

    :param array group_ids: ``(n,) dtype array``
    :param array values: ``(n,) float array``
    :param float max_gap: maximum distance between a value and the nearest value in the same group.
    :returns: ``(n,) uint array``,
        such that ``out[s[i]]==out[s[i+1]]`` :math:`\iff`
        ``group_ids[s[i]]==group_ids[s[i+1]]`` & ``|values[s[i]]-values[s[i+1]]| <= max_gap``
        where ``s[i]`` is the ``i`` -th index when sorting by id and value

    Example
    _______
    >>> group_ids = numpy.array([0, 0, 0, 0, 0, 1, 1, 1, 1, 3, 3, 3, 3])
    >>> values = numpy.array([ 0, 35, 20, 25, 30,  0,  5, 10, 20,  0,  5, 10, 15])
    >>> connect_adjacents_in_groups(group_ids, values, max_gap = 5)
    array([0, 1, 1, 1, 1, 2, 2, 2, 3, 4, 4, 4, 4], dtype=uint32)

    Example with string ``group_ids``:

    >>> group_ids = numpy.array(['alpha', 'alpha', 'alpha', 'alpha', 'alpha', 'beta', 'beta', 'beta', 'beta', 'gamma', 'gamma', 'gamma', 'gamma'])
    >>> values = numpy.array([ 0, 35, 20, 25, 30,  0,  5, 10, 20,  0,  5, 10, 15])
    >>> connected_components_ids = connect_adjacents_in_groups(group_ids, values, max_gap = 5)

    The function does not require the ``group_ids`` or the ``values`` to be sorted:

    >>> shuffler = numpy.random.permutation(len(group_ids))
    >>> group_ids_shuffled = group_ids[shuffler]
    >>> values_shuffled = values[shuffler]
    >>> connect_adjacents_in_groups(group_ids_shuffled, values_shuffled, max_gap = 5)
    array([2, 1, 0, 2, 4, 1, 1, 4, 1, 4, 3, 2, 4], dtype=uint32)
    >>> connected_components_ids[shuffler]
    array([2, 1, 0, 2, 4, 1, 1, 4, 1, 4, 3, 2, 4], dtype=uint32)
    """
    as_idx = False
    if group_ids.dtype.kind in 'ui':
        if min(group_ids) >= 0 and max(group_ids) < (1 << 6) * len(group_ids):
            as_idx = True
    # FIXME: add old max and old min for it to work with pandas DataFrames
    if as_idx:
        values_for_uint32 = linearscaling(
            values, 1, (1 << 32) - float(1 << 8) - 1)
        args = lexsort_uint32_pair(group_ids, values_for_uint32)
    else:
        args = numpy.lexsort((values, group_ids))
    group_ids = group_ids[args]  # e.g.  1 1 1 1 1 1 1 1 1 2 2 2 2
    values = values[args]  # e.g.        1 1 1 2 2 3 3 9 9 1 2 2 9
    # to_split  e.g.                     0 0 0 0 0 0 1 0 1 0 0 1
    to_split = ((group_ids[1:] != group_ids[:-1])
                | ((values[1:] - values[:-1]) > max_gap))
    # group_idx  e.g.                    0 0 0 0 0 0 0 1 1 2 2 2 3
    group_idx = numpy.empty(group_ids.size, dtype='uint32')
    group_idx[0] = 0
    numpy.cumsum(to_split, out=group_idx[1:])
    # reverse argsort
    aligned_group_idx = numpy.empty_like(group_idx)
    aligned_group_idx[args] = group_idx
    return aligned_group_idx


# TODO: improve test
def get_value_by_idx(idx, values, default, check_unique=True, minlength=None):
    """
    Given array of indexes ``idx`` and array ``values`` (unordered, not necesarilly full),
    output array such that ``out[i] = values[idx==i]``.

    If all indexes in ``idx`` are unique, it is equivalent to sorting the ``values``
    by their ``idx`` and filling  with ``default`` for missing ``idx``.

    If ``idx`` elements are not unique and you still want to proceed,
    you can set ``check_unique`` to ``False``. The output values for the non-unique indexes
    will be chosen arbitrarily among the multiple values corresponding.

    :param array idx: ``(n,) uint array`` with values < max_idx
    :param array values: ``(n,) dtype array``
    :param dtype default: filling value for ``output[i]`` if there is no ``idx == i``
    :param bool check_unique: if ``True``, will check that ``idx`` are unique
        If ``False``, if the ``idx`` are not unique, then an arbitrary value
        will be chosen.
    :param int? minlength: minimum shape for the output array (``default: idx.max() + 1``).
    :returns array: (max_idx+1,), dtype array such that
        ``out[i] = values[idx==i]``.

    Example
    _______
    >>> idx = numpy.array([8,2,4,7])
    >>> values = numpy.array([100, 200, 300, 400])
    >>> get_value_by_idx(idx, values, -1, check_unique=False, minlength=None)
    array([ -1,  -1, 200,  -1, 300,  -1,  -1, 400, 100])

    Example with non-unique elements in ``idx``:

    >>> idx = numpy.array([2,2,4,7])
    >>> values = numpy.array([100, 200, 300, 400])
    >>> get_value_by_idx(idx, values, -1, check_unique=False, minlength=None)
    array([ -1,  -1, 200,  -1, 300,  -1,  -1, 400])
    """
    assert idx.dtype.kind == 'u' or (idx.dtype.kind == 'i' and (idx >= 0).all()), (
        'Can only use get_xx_by_idx with integer indexes in `idx`, where (idx >= 0).all()')
    if check_unique:
        assert numpy.unique(idx).shape == idx.shape, "indexes in `idx` should be unique"
    length = max(idx.max() + 1, minlength or 0)
    out = numpy.full(length, default, dtype=values.dtype)
    out[idx] = values
    return out


# TODO: improve test and add example in doc
def get_most_common_by_idx(idx, values, fill, minlength=None):
    """
    Given array of indexes ``idx`` and array ``values``,
    outputs the most common value by idx.

    :param array idx: (n,) uint array with values < max_idx
    :param array values: (n,) non-float, dtype array
    :param fill: filling value for output[i] if there is no idx == i
    :param minlength: minimum shape for the output array.
    :returns: (max_idx+1,), dtype array such that
        out[i] = the most common value such that (values[idx==i])
    """
    assert idx.dtype.kind == 'u' or (idx.dtype.kind == 'i' and (idx >= 0).all()), (
        'Can only use get_xx_by_idx with integer idx, where all idx >= 0')
    assert values.dtype.kind != 'f', ('values dtype is float - Please convert to other dtype'
                                      'or digitize values before using get_most_common_by_idx')
    length = max(idx.max() + 1, minlength or 0)

    sessions_uv = to_structured([
        ('idx', idx),
        ('value', values),
    ])
    sessions_uv_idx, uv_idx2id, _ = factorize(sessions_uv)
    uv_idx2id = numpy.asarray(
        uv_idx2id, [('idx', 'uint32'), ('value', values.dtype)])  # cast to struct
    count_by_uv_idx = numpy.bincount(sessions_uv_idx)
    top_uv_by_id = argmax_by_idx(
        uv_idx2id['idx'], count_by_uv_idx, minlength=length)
    out = uv_idx2id[top_uv_by_id]['value']
    out[top_uv_by_id == -1] = fill
    return out


def average_by_idx(idx, values, weights=None, minlength=None, fill=0, dtype='float64'):
    """
    Compute average-by-idx given array of indexes ``idx``, ``values``, and optional ``weights``

    :param array idx: (n,) int array
    :param array values: (n,) float array
    :param array? weights: (n,) float array
    :param int? minlength: (default: idx.max() + 1)
    :param float? fill: filling value for missing idx (default: 0)
    :param str? dtype: (default: 'float32')
    :returns: (min-length,) float array, such that out[i] = mean(values[idx==i])

    Example
    _______
    >>> idx  = numpy.array([0, 0, 0, 0, 0, 1, 1, 1, 1, 3, 3, 3])
    >>> values   = numpy.array([0, 1, 2, 3, 4, 0, 2, 4, 6, 0, 4, 6])
    >>> average_by_idx(idx, values, fill=0)
    array([ 2.        ,  3.        , 0.        ,  3.33333333])
    >>> weights = numpy.array([0, 1, 0, 0, 0, 1, 2, 3, 4, 1, 1, 0])
    >>> average_by_idx(idx, values, weights=weights, fill=0)
    array([ 1.,  4., 0.,  2.])
    """
    assert idx.dtype.kind == 'u' or (idx.dtype.kind == 'i' and (idx >= 0).all()), (
        'Can only use get_xx_by_idx with integer idx, where (idx >= 0).all()')
    # FIXME: define dtype whitelist instead
    assert values.dtype.kind not in 'USOb', ('values dtype not supported')
    norm_by_idx = numpy.bincount(
        idx, weights, minlength=minlength).astype(dtype)
    if weights is not None:
        values = values * weights
    sum_by_idx = numpy.bincount(idx, values, minlength=minlength).astype(dtype)
    with numpy.warnings.catch_warnings():
        numpy.warnings.filterwarnings('ignore', r'.*divide.*')
        return numpy.where(norm_by_idx > 0, sum_by_idx / norm_by_idx, fill)



import sys

import logging

PYV = f'{sys.version_info.major}.{sys.version_info.minor}.{sys.version_info.micro}'


try:
    from tqdm import tqdm
except ImportError:
    def tqdm(seq, **kwargs):
        return seq


try:
    # if coloredlogs is installed and no logging handler is defined yet, configure it
    import coloredlogs
    if not logging.root.handlers:
        coloredlogs.install(fmt='%(asctime)s %(message)s', datefmt='%H:%M:%S')
except ImportError:
    logging.basicConfig()
logger = logging.getLogger()



"""
Array tools
===========
"""
import sys

import numpy

from xminds._lib.hashmap import Hashmap, factorize
from .arraybase import to_structured, set_or_add_to_structured


def structured_arrays_mean(arrays, keep_missing=False):
    """
    Computes average of each field given a list of struct-arrays.

    By default, will only average and return fields that are present in every struct-array in input.

    The average will be over concatenated fields as if they were in a single structured array,
    as opposed to averaging the results of several ``arr['field'].mean()``

    :param list-of-array arrays: list(struct-array)
    :param boolean? keep_missing: if True, use all fields from all arrays,
        including ones missing from some arrays;
        if False, only keep the intersection of fields defined in all arrays ``(default: False)``
    :returns: struct-array of shape (1,)

    Examples
    ________

    Obtaining the average of every field of a structured array:

    >>> arr1 = to_structured([
    >>>     ('a', numpy.arange(5)),
    >>>     ('b', 2 * numpy.arange(5))
    >>> ])
    >>> structured_arrays_mean([arr1])
    array([(2., 4.)],
      dtype=[('a', '<f8'), ('b', '<f8')])

    Getting the average of the same field scattered accross several structured arrays:

    >>> arr1 = to_structured([
    >>>     ('a', numpy.arange(5)),
    >>>     ('b', 2 * numpy.arange(5))
    >>> ])
    >>> arr2 = to_structured([
    >>>     ('a', numpy.ones(3)),
    >>>     ('b', 2 * numpy.ones(3)),
    >>>     ('c', 3 * numpy.ones(3))
    >>> ])
    >>> structured_arrays_mean([arr1, arr2])
    array([(1.625, 3.25)],
      dtype=[('a', '<f8'), ('b', '<f8')])

    If you want to include all fields, including fields that are only available in a few arrays,
    use option ``keep_missing=True``:

    >>> structured_arrays_mean([arr1, arr2], keep_missing=True)
    array([(1.625, 3.25, 3.)],
      dtype=[('a', '<f8'), ('b', '<f8'), ('c', '<f8')])
    """
    fields0 = set(arrays[0].dtype.names)
    if not keep_missing:
        fields_set = fields0.intersection(
            *(set(array.dtype.names) for array in arrays[1:]))
    else:
        fields_set = fields0.union(*(set(array.dtype.names)
                                     for array in arrays[1:]))
    return to_structured([
        (
            field,
            numpy.atleast_1d(numpy.concatenate(
                [array[field] for array in arrays if field in array.dtype.names]
            ).mean(axis=0, keepdims=True)))
        for field in fields_set
    ])


def first(a, predicate=None, batch_size=None, offset=0):
    """
    Efficiently find next index satisfying predicate in a 1d array.

    Can also be used on a array of booleans without a predicate.

    Will be added in numpy2.0: https://github.com/numpy/numpy/issues/2269

    :param array a: ``(n,) dtype array``
    :param callable? predicate: function(dtype-array -> bool-array) (or ``None`` if dtype=``bool``)
    :param int? batch_size: ``(default: 4k)``
    :param int? offset: ``(default: 0)``
    :returns: int, index of first value satisfying predicate
    :raises StopIteration: if there is no index satisfying predicate after ``offset``

    Examples
    ________

    With predicate:

    >>> a = numpy.array([0, 1, 1, 2, 3, 2, 4, 2])
    >>> idx = first(a, lambda x: x == 2)
    >>> idx
    3

    >>> idx = first(a, lambda x: x == 2, offset=idx + 1)
    >>> idx
    5

    >>> idx = first(a, lambda x: x == 2, offset=idx + 1)
    >>> idx
    7

    >>> idx = first(a, lambda x: x == 2, offset=idx + 1)
    StopIteration:

    Without predicate on array of booleans:

    >>> mask = numpy.array([0, 0, 0, 1, 0, 1, 0, 0], '?')
    >>> idx = first(mask)
    3
    """
    batch_size = batch_size or (1 << 12)
    n, = a.shape
    assert predicate is not None or a.dtype == '?', (
        'first without a predicate can only be used on an array of booleans')
    n_batches = int(numpy.ceil(float(n) / batch_size))
    for b in range(n_batches):
        batch_start = offset + b * batch_size
        batch_end = batch_start + batch_size
        batch = a[batch_start:batch_end]
        tests = batch if predicate is None else predicate(batch)
        inds = tests.nonzero()[0]
        try:
            return batch_start + inds[0]
        except IndexError:
            pass
    raise StopIteration()


def set_or_reallocate(array, values, offset, growing_factor=2., fill=None):
    """
    Assigns ``values`` (of length ``n2``) in ``array`` (of length ``n1``),
    starting at array's ``offset`` index.

    Returns the same array if there is enough space in array for assigning
    ``array[offset:offset+n2, :]`` = values and otherwise returns a new array,
    expanded by ``growing_factor`` when ``n2 > n1 - offset``

    :param array array: ``(n1, *extra-dims)`` array
    :param array values: ``(n2, *extra-dims)`` array
    :param int offset: index of ``array`` to start assigning ``values``.
              ``array[offset:offset+n2, :] = values``
    :param float? growing_factor: growing factor > 1 (``default:2``)
    :param float? fill: fill value or ``None`` to leave empty
    :returns: same array if ``n1 >= offset + n2``, or new array with copied data otherwise
              assigns ``array[offset:offset+n2, :] = values``

    Examples
    ________

    If ``n1 >= offset + n2``, the same array is returned with
    ``array[offset:offset+n2, :] = values``.
    In this case, the assignment is inplace and the input array will be affected.

    >>> array  = numpy.arange(10)
    >>> values = - 2 * numpy.arange(5)
    >>> set_or_reallocate(array, values, offset=5)
    array([ 0,  1,  2,  3,  4,  0, -2, -4, -6, -8])
    >>> array
    array([ 0,  1,  2,  3,  4,  0, -2, -4, -6, -8])

    Otherwise, a new expanded array with copied data is created and with
    ``array[offset:offset+n2, :] = values``.
    Without specifying a ``fill`` value, the newly expanded data will be arbitrary,
    starting at index ``offset + n2``.
    Since a new array is created, the input array will not be affected.

    >>> array  = numpy.arange(10)
    >>> set_or_reallocate(array, values, offset=10)
    array([                  0,                   1,                   2,
                             3,                   4,                   5,
                             6,                   7,                   8,
                             9,                   0,                  -2,
                            -4,                  -6,                  -8,
           5572452860762084442,     140512663005448,     670512663005448,
                             0,    2814751914590207])
    >>> array
    array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9])

    >>> set_or_reallocate(array, values, offset=10, fill=0)
    array([ 0,  1,  2,  3,  4,  5,  6,  7,  8,  9,  0, -2, -4, -6, -8,  0,  0,
        0,  0,  0])
    >>> array
    array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9])
    """
    values = numpy.asarray(values)
    end = offset + values.shape[0]
    n1 = array.shape[0]
    extra_dims = array.shape[1:]
    assert values.shape[1:] == extra_dims
    # Nothing to do if it fits
    if end <= n1:
        array[offset:end] = values
        return array
    # Grow if it does not fit
    if growing_factor <= 1:
        raise ValueError('growing_factor must be larger than 1')
    new_length = max(end, int(growing_factor * n1))
    dtype = array.dtype
    if fill is not None:
        new_array = numpy.full((new_length,) + extra_dims, fill, dtype=dtype)
    else:
        new_array = numpy.empty((new_length,) + extra_dims, dtype=dtype)
    # Fill previous values
    new_array[:n1] = array
    # Sets new values
    new_array[offset:end] = values
    return new_array


def lexsort_uint32_pair(a, b):
    """
    faster alternative to numpy.lexsort for two uint32

    :param array a: (n,) uint32-array
    :param array b: (n,) uint32-array
    :returns: permutation array of n to sort by a first and b second
    """
    assert a.min() >= 0 and a.max() < (1<<32), (a.min(), a.max())
    assert b.min() >= 0 and b.max() < (1<<32), (b.min(), b.max())
    assert a.size == b.size
    dtype = ([('b', 'u4'), ('a', 'u4')] if sys.byteorder == 'little'
             else [('a', 'u4'), ('b', 'u4')])
    combined = numpy.empty(a.size, dtype=dtype)
    combined['a'] = a
    combined['b'] = b
    return numpy.argsort(combined.view('u8'))


def kargmax(a, k, axis=0, do_sort=False):
    """
    Returns the indices of the k maximum values along a specified axis.
    Set ``do_sort`` at ``True`` to get these indices ordered to get ``a[output]`` sorted.
    User-friendly wrapper around numpy.argpartition

    :param arr a: (n,) input array
    :param int k: sets the k in 'top k maximum values'
    :param int? axis: axis along which to look for argmax
        Default: 0
    :param bool? do_sort: if False, then the order of the output will be arbitrary
        if True, then ``a[output]`` will be sorted, starting with the a.max().
        Default: False
    :return arr: (k,) indexes of the k maximum values along the specified axis

    Examples
    ________
    >>> a = numpy.array([3, 1, 9, 6, 4, 4, 0, 6, 4, 8, 1, 3])
    >>> kargmax(a, 2)
    array([9, 2])

    The output array's order is arbitrary - to have the output ordered so that ``a[output]``
    is sorted, then use ``do_sort=True``:

    >>> kargmax(a, 2, do_sort=True)
    array([2, 9])

    On ndarrays:

    >>> a = numpy.array([3, 1, 9, 5, 6, 4, 0, 6, 4, 8, 1, 3]).reshape((3, 4))
    array([[3, 1, 9, 5],
           [6, 4, 0, 6],
           [4, 8, 1, 3]])

    >>> kargmax(a, 2, axis=0)
    array([[2, 1, 2, 0],
           [1, 2, 0, 1]])

    >>> kargmax(a, 2, axis=1)
    array([[3, 2],
           [0, 3],
           [0, 1]])
    """
    a = numpy.asarray(a)
    ndim = len(a.shape)
    n = a.shape[axis]
    if n <= k:
        topk = numpy.indices(a.shape)[axis]
    else:
        argp = numpy.argpartition(a, -k, axis=axis)
        if axis == 0:
            topk = argp[-k:]
        elif axis == 1:
            topk = argp[:, -k:]
        else:
            raise NotImplementedError()
    if not do_sort:
        return topk
    if ndim == 1:
        args = numpy.argsort(-a[topk], axis=axis)
        return topk[args]
    elif ndim == 2:
        if axis == 0:
            cols = numpy.indices(topk.shape)[1]
            args = numpy.argsort(-a[topk, cols], axis=0)
            return topk[args, cols]
        else:
            assert axis == 1
            rows = numpy.indices(topk.shape)[0]
            args = numpy.argsort(-a[rows, topk], axis=1)
            return topk[rows, args]
    raise NotImplementedError(
        '`kargmax` with `do_sort` only supports 1d or 2d')


def kargmin(a, k, axis=0, do_sort=False):
    """
    Returns the indices of the k minimum values along a specified axis.
    Set ``do_sort`` at ``True`` to get these indices ordered to get a[output] sorted.
    User-friendly wrapper around numpy.argpartition.

    :param arr a: (n,) input array
    :param int k: sets the k in 'top k minimum values'
    :param int? axis: axis along which to look for argmin
        Default: 0
    :param bool? do_sort: if False, then the order of the output will be arbitrary
        if True, then ``a[output]`` will be sorted, starting with the a.min().
        Default: False
    :return arr: (k,) indexes of the k minimum values along the specified axis

    Examples
    ________
    >>> a = numpy.array([3, 1, 9, 6, 4, 4, 0, 6, 4, 8, 1, 3])
    >>> kargmin(a, 2)
    array([1, 6])

    The output array's order is arbitrary - to have the output ordered so that ``a[output]``
    is sorted, then use ``do_sort=True``:

    >>> kargmin(a, 2)
    array([6, 1])

    On ndarrays:

    >>> a = numpy.array([3, 1, 9, 5, 6, 4, 0, 6, 4, 8, 1, 3]).reshape((3, 4))
    array([[3, 1, 9, 5],
           [6, 4, 0, 6],
           [4, 8, 1, 3]])

    >>> kargmin(a, 2, axis=0)
    array([[0, 0, 1, 2],
           [2, 1, 2, 0]])

    >>> kargmin(a, 2, axis=1)
    array([[1, 0],
           [2, 1],
           [2, 3]])


    """
    a = numpy.asarray(a)
    ndim = len(a.shape)
    n = a.shape[axis]
    if n <= k:
        topk = numpy.indices(a.shape)[axis]
    else:
        argp = numpy.argpartition(a, k, axis=axis)
        if axis == 0:
            topk = argp[:k]
        elif axis == 1:
            topk = argp[:, :k]
        else:
            raise NotImplementedError()
    if not do_sort:
        return topk
    if ndim == 1:
        args = numpy.argsort(a[topk], axis=axis)
        return topk[args]
    elif ndim == 2:
        if axis == 0:
            cols = numpy.indices(topk.shape)[1]
            args = numpy.argsort(a[topk, cols], axis=0)
            return topk[args, cols]
        else:
            assert axis == 1
            rows = numpy.indices(topk.shape)[0]
            args = numpy.argsort(a[rows, topk], axis=1)
            return topk[rows, args]
    raise NotImplementedError(
        '`kargmin` with `do_sort` only supports 1d or 2d')


def in1d(needles, haystack, invert=False):
    """
    Test whether each element of a 1d array is also present in a second 1d array.

    :param array needles: (n1,) dtype array
    :param array haystack: (n2,) dtype array
    :param bool? invert: if True, the values in the returned array are inverted,
        ``in1d(a, b, invert=True)`` is equivalent to ``~in1d(a, b)``. (``default: False``)
    :returns: (n1,) bool array

    Example
    _______
    >>> needles = numpy.array([5,10,20])
    >>> haystack = numpy.arange(15)
    >>> in1d(needles, haystack)
    array([True, True, False])
    >>> in1d(needles, haystack, invert=True)
    array([False, False, True])
    """
    if invert:
        return ~Hashmap(haystack).contains(needles)
    return Hashmap(haystack).contains(needles)


def search(needles, haystack, idx_dtype='uint32'):
    """
    Return whether each element of a 1d array is present in a second 1d array
    and the corresponding indexes.

    If an element of ``needles`` is not found in ``haystack``, the corresponding value
    returned in ``indexes`` is 0.

    :param array needles: (n1,) dtype array
    :param array haystack: (n2,) dtype array
    :param dtype? idx_dtype: (``default: uint32``)
    :returns: tuple(
        indexes: (n1,) idx_dtype array <n2 of indexes in haystack,
        found: (n1,) bool array,)

    Example
    _______
    >>> needles = numpy.array([1000, 2000, 3000])
    >>> haystack = numpy.arange(50)
    >>> haystack[10] = needles[0]
    >>> haystack[20] = needles[1]
    >>> search(needles, haystack)
    (array([10, 20,  0], dtype=uint32), array([ True,  True, False]))
    """
    hashmap = Hashmap(haystack, numpy.arange(haystack.size, dtype=idx_dtype))
    idx, found = hashmap.get_many(needles)
    return idx, found


def cumcount_by_value(values, assume_sorted=False):
    """
    Compute the rank of appearance of entries grouped by input value.

    If a value appears for the 5th time in the ``values`` array at index ``i``,
    ``output[i]`` will be a 5.

    :param array values: uint32 array
    :param bool? assume_sorted: if True, the input values are assumed sorted (``default: False``)
        for non-integer values, the sort order can be arbitrary (see examples below)
        for integers values, the order must be the one of increasing integers
    :returns: array of int

    Example
    _______
    >>> array = numpy.array([0, 0, 0, 0, 1, 1, 1, 3, 3, 3, 3, 3])
    >>> cumcount_by_value(array)
    array([0, 1, 2, 3, 0, 1, 2, 0, 1, 2, 3, 4])
    >>> array = numpy.array(['beta', 'alpha', 'gamma', 'alpha', 'beta', 'alpha', 'delta'])
    >>> cumcount_by_value(array)
    array([0, 0, 0, 1, 1, 2, 0])
    >>> array = numpy.array(['A', 'B', 'B', 'C', 'C', 'C'])
    >>> cumcount_by_value(array, assume_sorted=True)
    array([0, 0, 1, 0, 1, 2])

    When using ``assume_sorted`` on non-integer values, the order can be arbitrary:

    >>> ['B', 'B', 'C', 'C', 'C', 'A']
    >>> cumcount_by_value(array, assume_sorted=True)
    array([0, 1, 0, 1, 2, 0])
    """
    if len(values) == 0:
        return []
    convert = True
    if values.dtype.kind in 'ui':
        if min(values) >= 0 and max(values) < (1 << 6) * len(values):
            convert = False
    if convert and assume_sorted:
        values = numpy.asarray(values)
        where_diff, = numpy.where(values[1:] != values[:-1])
        n_per_value = numpy.r_[-1, where_diff, values.size - 1]
        n_per_value[1:] -= n_per_value[:-1]
        return _arange_sequence(n_per_value[1:])
    if convert:
        values, _, _ = factorize(values)
    if not assume_sorted:
        argsort = numpy.argsort(values, kind='stable')
    # values        0 0 0 0 1 1 1 3 3 3 3 3
    # n_per_idx     4 3 0 5
    n_per_idx = numpy.bincount(values)
    # rank          0 1 2 3 0 1 2 0 1 2 3 4
    rank = _arange_sequence(n_per_idx)
    if assume_sorted:
        return rank
    # reverse argsort
    out = numpy.empty_like(rank)
    out[argsort] = rank
    return out


def _arange_sequence(lengths):
    """
    :param array lengths: int array like [4 3 0 5]
    :returns: concat(arange(i) for i length) like [0 1 2 3 0 1 2 0 1 2 3 4]
    """
    # lengths     4 3 0 5
    # cum_counts  0 4 7 7 12
    cum_counts = numpy.zeros(lengths.size + 1, dtype=lengths.dtype)
    numpy.cumsum(lengths, out=cum_counts[1:])
    # offset      0 0 0 0 4 4 4 7 7 7 7 7
    offset = numpy.repeat(cum_counts[:-1], lengths)
    # out         0 1 2 3 0 1 2 0 1 2 3 4
    return numpy.arange(offset.size) - offset


def unique_count(values):
    """
    Count the number of unique elements in ``values``.

    :param array values: (n,) dtype array
    :returns: int number of unique values
    """
    return Hashmap(keys=values, values=None).n_used



"""
Iterable tools
==============
"""
from itertools import chain, islice


def split(iterable, function):
    """
    Split an iterable into two lists according to test function

    :param iterable iterable: iterable of values to be split
    :param function function: decision function ``value => bool``
    :returns: tuple(
        list with values for which function is `True`,
        list with values for which function is `False`,)

    Example
    _______
    >>> split([1,2,3,4,5,6], lambda x: x<3)
    ([1, 2], [3, 4, 5, 6])
    """
    match = []
    unmatch = []
    for value in iterable:
        if function(value):
            match.append(value)
        else:
            unmatch.append(value)
    return match, unmatch


def unzip(zipped):
    """
    Unzip a zipped list

    :param list-of-tuple zipped: list of tuples to be disaggregated
    :returns: list of tuples

    Example
    _______

    >>> unzip([(1, 2, 3), (4, 5, 6), (7, 8, 9)])
    [(1, 4, 7), (2, 5, 8), (3, 6, 9)]
    """
    return list(zip(*zipped))


def is_empty_iterable(iterable):
    """
    Check if a sequence is empty.

    Most useful on generator types.

    :param iterable iterable: input iterable
    :returns: tuple(iterable, is_empty). If a generator is passed,
        a new generator will be returned preserving the original values

    Example
    _______
    >>> a = []
    >>> b = (str(i) for i in range(0))
    >>> c = (str(i) for i in range(5))

    >>> a, is_empty = is_empty_iterable(a)
    >>> a, is_empty
    ([], True)
    >>> b, is_empty = is_empty_iterable(b)
    >>> is_empty
    True

    When the generator ``c`` is given, a new generator is returned by ``is_empty_iterable``
    to preserve original values of ``c``:

    >>> c, is_empty = is_empty_iterable(c)
    >>> next(c), is_empty
    ('0', False)
    """
    if hasattr(iterable, '__len__'):
        is_empty = len(iterable) == 0
        return iterable, is_empty
    # iterable is likely a generator-like, we need to consume one item and re-build a new generator
    iterable = iter(iterable)
    try:
        first = next(iterable)
    except StopIteration:
        return iterable, True
    return chain([first], iterable), False


def get_first_of_iterable(iterable):
    """
    Return the first element of the given sequence.

    Most useful on generator types.

    :param iterable iterable: input iterable
    :returns: tuple(iterable, first_element). If a generator is passed,
        a new generator will be returned preserving the original values.

    :raises: IndexError

    Example
    _______
    >>> a = [1,2,3]
    >>> b = (str(i) for i in range(3))

    >>> a, first_element = get_first_of_iterable(a)
    >>> a, first_element
    ([1, 2, 3], 1)

    When the generator ``b`` is given, a new generator is returned by ``is_empty_iterable``
    to preserve original values of ``b``:

    >>> b, first_element = get_first_of_iterable(b)
    >>> next(b), first_element
    ('0', '0')
    """
    if hasattr(iterable, '__getitem__'):
        return iterable, iterable[0]
    iterable = iter(iterable)
    try:
        first = next(iterable)
    except StopIteration:
        raise IndexError('`iterable` is empty')
    return chain([first], iterable), first


def ichunk_iterable(iterable, chunk_length=2**12):
    """
    Split a sequence into consecutive sub-sequences of given length by returning
    a generator of generators generating the sub sequences.

    Most useful on generator types.

    :param iterable iterable: input iterable
    :param int? chunk_length: length of the chunks. (``default: 4096``)
    :generates: generators

    Example
    _______
    >>> iterable = range(15)
    >>> chunks = ichunk_iterable(iterable, 4)
    >>> for chunk in chunks:
    >>>    print(list(chunk))
    [0, 1, 2, 3]
    [4, 5, 6, 7]
    [8, 9, 10, 11]
    [12, 13, 14]
    """
    iterable = iter(iterable)
    for head in iterable:
        def chunk():
            yield head
            for val in islice(iterable, chunk_length - 1):
                yield val
        yield chunk()



"""
Arrays base tools
=================
"""
import numpy
from numpy.lib.recfunctions import unstructured_to_structured

from .iterable import split


def to_structured(arrays):
    """
    Casts list of array and names (and optional dtypes) to numpy structured array.

    See `Numpy's documentation <https://numpy.org/doc/stable/user/basics.rec.html>`_
    for how to use numpy's structured arrays.

    A pandas DataFrame can be converted into a numpy record array,
    using DataFrame.to_records() (`to_records documentation
    <https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.to_records.html>`_,
    `documentation on record arrays
    <https://numpy.org/doc/stable/user/basics.rec.html#record-arrays>`_).
    A record array can then be converted into a structured array by using:

    >>> recordarr.view(recordarr.dtype.fields, numpy.ndarray)

    :param list-of-tuple arrays: ``list((name, array-like, *dtype-infos)))``
        or ``{name: array-like}``.
        A lone value will be broadcasted as an array full of this value
        and of the size of the other arrays
    :returns: struct-array

    Examples
    ________
    >>> to_structured([
    >>>     ('a', numpy.arange(5), 'uint32'),
    >>>     ('b', 2 * numpy.arange(5), 'float32')
    >>> ])
    array([(0, 0.), (1, 2.), (2, 4.), (3, 6.), (4, 8.)],
      dtype=[('a', '<u4'), ('b', '<f4')])

    A single value can also be used, and will be broadcasted as an array full of this value,
    and of the size of the other arrays, for instance:

    >>> to_structured([
    >>>     ('a', numpy.arange(5), 'uint32'),
    >>>     ('b', 2, 'float32')
    >>> ])
    array([(0, 2.), (1, 2.), (2, 2.), (3, 2.), (4, 2.)],
      dtype=[('a', '<u4'), ('b', '<f4')])

    Not specifying the dtype in the tuples will cause the function to use the array's dtype,
    or to infer it in case of sequences of Python objects.

    >>> to_structured([
    >>>    ('a', numpy.arange(5, dtype='uint32')),
    >>>    ('b', [2 * i for i in range(5)])
    >>> ])
    array([(0, 0), (1, 2), (2, 4), (3, 6), (4, 8)],
      dtype=[('a', '<u4'), ('b', '<i8')])

    Using dictionaries:

    >>> to_structured({
    >>>     'a': numpy.arange(5, dtype='uint32'),
    >>>     'b': [2 * i for i in range(5)]
    >>> })
    array([(0, 0), (1, 2), (2, 4), (3, 6), (4, 8)],
      dtype=[('a', '<u4'), ('b', '<i8')])

    (n, m) 2D numpy arrays are viewed as (n,) arrays of (m,) 1D arrays:

    >>> to_structured([
    >>> ('a', numpy.arange(5)),
    >>> ('b', numpy.arange(15).reshape(5, 3))
    >>> ])
    array([(0, [ 0,  1,  2]), (1, [ 3,  4,  5]), (2, [ 6,  7,  8]),
       (3, [ 9, 10, 11]), (4, [12, 13, 14])],
      dtype=[('a', '<i8'), ('b', '<i8', (3,))])
    """
    if isinstance(arrays, dict):
        arrays = arrays.items()
    else:
        assert isinstance(arrays, list), (
            'arrays must be a dict or a list, not {}'.format(type(arrays)))
    all_n_rows = {len(a) for _, a, *_ in arrays if numpy.asarray(a).shape}
    if len(all_n_rows) > 1:
        raise ValueError(
            'all arrays must have the same shape[0], got: {}'.format(all_n_rows))
    n_rows, = all_n_rows
    # set default dtype
    dtypes = []
    for t in arrays:
        name, array = t[:2]
        if len(t) >= 3:  # given dtype infos
            dtype_infos = t[2:]
        else:  # infer dtype and shape
            array = numpy.asarray(array)
            if len(array.shape) == 1:
                dtype_infos = (array.dtype,)
            else:
                # do not use e.g. ('int', 3) but always '(3,)int' because the former flattens '1'
                dtype_infos = (f'{array.shape[1:]}{array.dtype}',)
        dtypes.append((name,) + dtype_infos)
    # combine to structured-array
    out = numpy.empty(n_rows, dtype=dtypes)
    for t in arrays:
        out[t[0]] = t[1]
    return out


def set_or_add_to_structured(array, data, copy=True):
    """
    Updates existing structured array, either by replacing the data for existing fields,
    or by adding new fields to the array.

    Fast `alternative <https://github.com/numpy/numpy/issues/7811>`_ to
    ``numpy.lib.recfunctions.append_fields``.

    :param struct-array array: array to update
    :param list-of-tuple data: list((name, array-or-scalar))  (scalars are broadcasted)
    :param bool? copy: set to False to avoid copy when possible ``(default: True)``
    :returns: struct-array

    Examples
    ________
    Adding field to existing structured array:

    >>> array = to_structured([
    >>>     ('a', numpy.arange(5, dtype='uint8')),
    >>>     ('b', 2 * numpy.arange(5, dtype='uint16')),
    >>> ])
    >>> new_data = 3 * numpy.arange(5, dtype='float32')
    >>> updated_array = set_or_add_to_structured(array, [
    >>>     ('c', new_data),
    >>> ])
    >>> updated_array
    array([(0, 0,  0.), (1, 2,  3.), (2, 4,  6.), (3, 6,  9.), (4, 8, 12.)],
      dtype=[('a', 'u1'), ('b', '<u2'), ('c', '<f4')])

    Replacing data from a structured array:

    >>> updated_array = set_or_add_to_structured(array, [
    >>>     ('b', new_data)
    >>> ])
    array([(0,  0), (1,  3), (2,  6), (3,  9), (4, 12)],
      dtype=[('a', 'u1'), ('b', '<u2')])

    Or doing both, while adding broadcasted constants:

    >>> updated_array = set_or_add_to_structured(array, [
    >>>     ('b', new_data),
    >>>     ('c', 2 * new_data),
    >>>     ('d', 1),
    >>>     ('e', b'1')
    >>> ])
    array([
        (0,  0,  0., 1, b'1'),
        (1,  3,  6., 1, b'1'),
        (2,  6, 12., 1, b'1'),
        (3,  9, 18., 1, b'1'),
        (4, 12, 24., 1, b'1')],
        dtype=[('a', 'u1'), ('b', '<u2'), ('c', '<f4'), ('d', '<i8'), ('e', 'S1')])
    """
    existing, new = split(data, lambda t: t[0] in array.dtype.names)
    if existing:
        if copy:
            array = array.copy()
        for k, v in existing:
            array[k] = v
    if new:
        arrays = [array]
        for name, data in new:
            data = numpy.atleast_1d(data)
            if len(data.shape) == 1:
                data = data.astype([(name, data.dtype)])
            else:
                # pack 2d arrays into structure
                dtype = numpy.dtype([(name, data.dtype, data.shape[1:])])
                data = unstructured_to_structured(data, dtype)
            arrays.append(data)
        all_dtypes = [
            (name, a.dtype.fields[name][0])
            for a in arrays
            for name in a.dtype.names
        ]
        # ``merge_arrays`` is really slow, so we copy manually
        array = numpy.empty(array.size, dtype=all_dtypes)
        for a in arrays:
            for name in a.dtype.names:
                array[name] = a[name]
    return array


def clean_dtype(dtype, sort=False):
    """
    Remove offsets from dtype, keeping only names and dtype.
    (See `Numpy dtype documentation <https://numpy.org/doc/stable/reference/generated/numpy.dtype.html>`_.)

    :param dtype-descr dtype: either a numpy.dtype, or a description of it
    :param bool? sort: (default: False)
    :returns: clean dtype, without offsets, sorted by field-names

    Example
    _______
    >>> d = numpy.dtype({
    >>>    'names': ['z_col', 'd_col', 'a_col'],
    >>>    'formats': ['i4', 'f4','i4'],
    >>>    'offsets': [0, 4, 40]
    >>> })
    >>> d
    dtype({'names':['z_col','d_col','a_col'], 'formats':['<i4','<f4','<i4'], 'offsets':[0,4,40], 'itemsize':44})
    >>> clean_dtype(d)
    [('a_col', dtype('int32')),
    ('d_col', dtype('float32')),
    ('z_col', dtype('int32'))]
    """
    dtype = numpy.dtype(dtype)
    if not getattr(dtype, 'names', None):
        return dtype
    dtype = ((field, dtype) for field, (dtype, _) in dtype.fields.items())
    if sort:
        dtype = sorted(dtype)
    else:
        dtype = list(dtype)
    return numpy.dtype(dtype)


def remove_structured_offset(array):
    """
    Remove offset fields from structured array.
    Does not copy the data if the dtype does not have offsets.

    :param array array: structured array
    :returns: structured array without offsets

    Example
    _______
    >>> a = numpy.array([(1, 2, 3), (4, 5, 6)], [('a', 'i4'), ('b', 'i4'), ('c', 'i4')])
    >>> b = a[['c', 'a']]
    >>> b.dtype
    dtype({'names':['c','a'], 'formats':['<i4','<i4'], 'offsets':[8,0], 'itemsize':12})
    >>> b = remove_structured_offset(b)
    >>> b.dtype
    dtype([('c', '<i4'), ('a', '<i4')])
    """
    cleaned_dtype = clean_dtype(array.dtype, sort=False)
    return array.astype(cleaned_dtype)



""" Crossing Minds data science python library and API client

See our [readthedocs](https://xminds.readthedocs.io/en/latest/) for the Python Data Science library documentation.

See the [API Documentation](https://docs.api.crossingminds.com/) for the Crossing Minds universal recommendation API documentation.

Our tools are licensed under the MIT License.
"""
DOCLINES = (__doc__ or '').split("\n")

import setuptools

from xminds import __version__

setuptools.setup(
    name='xminds',
    version=__version__,
    author='Crossing Minds, Inc',
    author_email='contact@crossingminds.com',
    description=DOCLINES[0],
    long_description="\n".join(DOCLINES[2:]),
    long_description_content_type='text/markdown',
    url='https://github.com/Crossing-Minds/xminds-python',
    packages=['xminds', 'xminds.api', 'xminds.lib', 'xminds.ds', 'xminds._lib'],
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.6',
    install_requires=[
        'requests>=2.19.1',
        'numpy>=1.16',
    ],
)



"""
Scaling tools
===================
"""
import numpy


def linearscaling(x, new_min, new_max, old_min=None, old_max=None, axis=None):
    """
    Linearly rescale input from its original range to a new range.

    :param scalar-or-array x:  scalar or arrays of scalars in ``[old_min, old_max]`` of shape ``(n, *shape)``
    :param scalar-or-array new_min: scalar or array of shape ``(*shape,)``
    :param scalar-or-array new_max: scalar or array of shape ``(*shape,)``
    :param scalar-or-array? old_min: (``default=x.min()``)
    :param scalar-or-array? old_max: (``default=x.max()``)
    :param int? axis: (``default=None``)
    :return: scalar or array of scalars in ``[new_min, new_max]`` of shape ``(n, *shape)``

    Example
    _______
    >>> linearscaling(0, -10, 10, 0, 1)
    -10.0

    When the original range is not passed on, it is considered to be the interval in which the input values are.

    >>> x = numpy.arange(0, 1, 0.1)
    >>> linearscaling(x, 0, 10)
    array([0., 1., 2., 3., 4., 5., 6., 7., 8., 9.])
    >>> linearscaling(x, 0, 10, 0, 2)
    array([0. , 0.5, 1. , 1.5, 2. , 2.5, 3. , 3.5, 4. , 4.5])

    Linear scaling can be performed on dfferent series of data having different ranges provided that the new minima/maxima are specfied for each serie.

    >>> new_min = numpy.array([0, 0])
    >>> new_max = numpy.array([1, 100])
    >>> old_min = numpy.array([0, 0])
    >>> old_max = numpy.array([10, 10])
    >>> x = numpy.arange(0, 1, 0.1).reshape(5, 2)
    >>> x
    array([[0. , 0.1],
       [0.2, 0.3],
       [0.4, 0.5],
       [0.6, 0.7],
       [0.8, 0.9]])
    >>> linearscaling(x, new_min, new_max, old_min, old_max)
    array([[ 0.1, 10. ],
       [ 0.2, 20. ],
       [ 0.3, 30. ],
       [ 0.4, 40. ],
       [ 0.5, 50. ]])
    """
    if not isinstance(x, numpy.ndarray):
        # meant for scalars
        assert old_min is not None
        assert old_max is not None
        assert old_min <= x <= old_max
    else:
        if x.size == 0:
            return x
        # set default old to current
        if old_min is None:
            old_min = x.min(axis=axis)
        if old_max is None:
            old_max = x.max(axis=axis)
        if axis is not None:
            old_min = numpy.expand_dims(old_min, axis)
        if axis is not None:
            old_max = numpy.expand_dims(old_max, axis)
        tol = abs(x.max()) * 1e-06  # for float32 precision
        assert (old_min - tol <= x).all(axis=None) & (x <= old_max + tol).all(axis=None), (
            x.min(), x.max(), old_max, old_min)
    # clipping because we allow a tolerance in the previous assert
    return numpy.clip(
        new_min + (x - old_min) * (new_max - new_min) / (old_max - old_min + 1e-32),
        a_min=numpy.minimum(new_min, new_max),  # we support new_max < new_min, for inversions
        a_max=numpy.maximum(new_max, new_min)
    )
